from __future__ import annotations

import re
import uuid
from os import path
from typing import TYPE_CHECKING, Dict, List, Optional

from ..logger import logger
from .backend import Backend, Partition, SaveMode
from .backend import Table as BackendTable
from .backend import TableMeta as Table
from .common import SqlProcessorException

if TYPE_CHECKING:
    from .context import ProcessorContext
    from .funcs import FuncRunner

__all__ = ["StepConfig", "Step", "StepType", "ReportCollector", "StepFactory"]


class ReportCollector:
    def collect_report(self, step: Step, status: str = None, message: str = None):
        raise NotImplementedError()


class StepConfig:
    STEP_CONFIG_PATTERN = r"^-- target\s*=\s*(\w+)(.*)$"

    def __init__(self, step_type: str, step_name: str, condition: str, line_no: int, step_config_str: str = ""):
        self.step_type = step_type
        self.name = step_name
        self.condition = condition
        self.line_no = line_no
        self.step_config_str = step_config_str

    def __str__(self):
        return f"StepConfig(target={self.step_type}.{self.name}, condition={self.condition}, line_no={self.line_no})"

    def __repr__(self):
        return str(self)

    def __eq__(self, other: StepConfig):
        return (
            isinstance(other, StepConfig)
            and self.step_type == other.step_type
            and self.name == other.name
            and self.condition == other.condition
            and self.line_no == other.line_no
        )

    @staticmethod
    def from_config_line(config_line: str, line_no: int) -> StepConfig:
        configs = re.compile(r"^\s*-- ").sub("", config_line).strip()
        configs = configs[configs.index("=") + 1 :]
        target_type = configs[: configs.index(".")] if configs.find(".") != -1 else configs
        target_type = target_type[: target_type.index(",")] if target_type.find(",") != -1 else target_type
        if target_type not in StepType.all():
            raise SqlProcessorException(f"unknown step type: type={target_type}, supported_types={StepType.all()}")

        target_name = configs[configs.index(".") + 1 :] if configs.find(".") != -1 else None
        target_name = target_name.strip() if target_name is not None and target_name.strip() != "" else None
        target_condition = None

        if target_name is not None:
            condition_match = re.compile(r"^\s*(.*)\s*,\s*if\s*=(.*$)").match(target_name)
        else:
            condition_match = re.compile(r"^\s*(,)\s*if\s*=(.*$)").match(configs[len(target_type) :])
        if condition_match:
            target_name = condition_match.group(1) if target_name is not None else None
            target_condition = condition_match.group(2).strip()
            reg_exp = r"[a-zA-Z0-9_]*\([^()]*\)"
            if not re.compile(reg_exp).match(target_condition):
                raise SqlProcessorException(
                    f"parse step config failed. condition must be like {reg_exp}, "
                    f"but got {target_condition}. config_line={config_line}"
                )

        return StepConfig(
            step_name=target_name,
            step_type=target_type,
            condition=target_condition,
            line_no=line_no,
            step_config_str=config_line,
        )

    def is_target_name_a_func(self):
        return "(" in self.name if self.name is not None else False

    def has_condition(self):
        return self.condition is not None


class StepType:
    TEMP = "temp"
    CACHE = "cache"
    BROADCAST = "broadcast"
    TEMPLATE = "template"
    FUNC = "func"
    LOG = "log"
    CHECK = "check"
    VARIABLES = "variables"
    LIST_VARIABLES = "list_variables"
    HIVE = "hive"
    OUTPUT = "output"
    ACTION = "action"

    @staticmethod
    def all() -> List[str]:
        return [
            StepType.TEMP,
            StepType.CACHE,
            StepType.TEMPLATE,
            StepType.LOG,
            StepType.CHECK,
            StepType.VARIABLES,
            StepType.HIVE,
            StepType.FUNC,
            StepType.BROADCAST,
            StepType.OUTPUT,
            StepType.LIST_VARIABLES,
            StepType.ACTION,
        ]


class Step:
    def __init__(
        self,
        id: str,
        reporter_collector: ReportCollector,
        func_runner: FuncRunner,
        target_config: StepConfig = None,
        select_sql: str = None,
        debug_var_tmpl_replace: bool = False,
    ):
        self.id = id
        self.target_config = target_config
        self.select_sql = select_sql
        self.debug_var_tmpl_replace = debug_var_tmpl_replace
        self.reporter_collector = reporter_collector
        self.func_runner = func_runner

    def __str__(self):
        return str(self.target_config).replace("StepConfig(", "Step(", 1)

    def __repr__(self):
        return f"[\n    config: {self.target_config},\n    sql: {self.select_sql}\n]"

    def should_run(self, context: ProcessorContext):
        variables: dict = context.vars_context.vars
        if "__skip_all__" in variables and variables["__skip_all__"] in ["TRUE", True, 1, "True", "true", "1"]:
            return False
        if not self.target_config.has_condition():
            return True
        return self.func_runner.run_func(self.target_config.condition, context.vars_context)

    def read(self, backend: Backend, context: ProcessorContext) -> Optional[BackendTable]:
        if self.target_config.step_type in [StepType.TEMPLATE] or (
            self.target_config.step_type == StepType.CHECK and self._should_skip_check(context.vars_context.vars)
        ):
            return backend.create_empty_table()
        if self.target_config.is_target_name_a_func():
            if self.select_sql:
                self.preprocess_select_sql(context)
            return backend.create_empty_table()
        self.preprocess_select_sql(context)
        if self.target_config.step_type == StepType.ACTION:
            backend.exec_native_sql(self.select_sql)
            return None
        else:
            return backend.exec_sql(self.select_sql)

    def preprocess_select_sql(self, context):
        self.select_sql = context.replace_templates(self.select_sql)
        self.select_sql = context.replace_variables(self.select_sql)

    def write(self, backend: Backend, table: Optional[BackendTable], context: ProcessorContext, dry_run: bool = False):
        variables: dict = context.vars_context.vars

        if not table:
            return

        if StepType.VARIABLES == self.target_config.step_type and not table.is_empty():
            field_names = table.field_names()
            row = table.first()
            for field_name in field_names:
                index = field_names.index(field_name)
                field_value = "null"
                if row[index] is not None:
                    field_value = str(row[index])
                context.add_vars({field_name: field_value})

        if StepType.LIST_VARIABLES == self.target_config.step_type:
            field_names = table.field_names()
            rows = table.collect()
            list_vars = {}
            for field_name in field_names:
                index = field_names.index(field_name)
                list_vars[field_name] = [row[index] for row in rows]
            context.add_list_vars(list_vars)

        if StepType.TEMPLATE == self.target_config.step_type:
            context.add_templates({self.target_config.name: self.select_sql})

        elif StepType.TEMP == self.target_config.step_type:
            backend.create_temp_table(table, self.target_config.name)

        elif StepType.CACHE == self.target_config.step_type:
            if "__no_cache__" in variables and variables["__no_cache__"] in ["TRUE", True, 1, "True", "true"]:
                backend.create_temp_table(table, self.target_config.name)
            else:
                backend.create_cache_table(table, self.target_config.name)

        elif StepType.BROADCAST == self.target_config.step_type:
            backend.broadcast_table(table, self.target_config.name)

        elif StepType.LOG == self.target_config.step_type:
            if "__no_log__" in variables and variables["__no_log__"] in ["TRUE", True, 1, "True", "true"]:
                return
            self._write_for_log_step(table)

        elif StepType.FUNC == self.target_config.step_type:
            self.func_runner.run_func(self.target_config.name, context.vars_context)

        elif StepType.CHECK == self.target_config.step_type:
            if self._should_skip_check(variables):
                return
            self._write_for_check_step(table, context)

        elif self.target_config.step_type in [StepType.HIVE, StepType.OUTPUT]:
            self._write_for_output_step(backend, table, context, dry_run)

    def _should_skip_check(self, variables):
        return "__no_check__" in variables and variables["__no_check__"] in ["TRUE", True, 1, "True", "true"]

    def is_template_statement(self):
        return StepType.TEMPLATE == self.target_config.step_type

    def _write_for_output_step(self, backend: Backend, table: BackendTable, context: ProcessorContext, dry_run: bool):
        extra_cols, variables = context.extra_cols, context.vars
        if "." not in self.target_config.name:
            message = (
                "table name for hive or output must be a full name, it should be of format DB.TABLE_NAME, got"
                f" `{self.target_config.name}`"
            )
            self.collect_report(message=message)
            raise SqlProcessorException(message)

        temp_table_name = f'{self.target_config.name.split(".")[1]}_{uuid.uuid4().hex}'
        for col in extra_cols:
            table = table.with_column(col.name, col.value)
        backend.create_temp_table(table, temp_table_name)

        source_table = Table(temp_table_name)
        target_table_name = f"{self.target_config.name}"

        static_partition_name, static_partition_value, create_output_table, save_mode = (
            None,
            None,
            False,
            SaveMode.overwrite,
        )
        for name, value in variables.items():
            if "__partition__" in name:
                static_partition_name = name[len("__partition__") :]
                if backend.is_spark_backend:
                    static_partition_value = value
                else:
                    static_partition_value = backend.sql_expr.convert_partition_value(static_partition_name, value)
            if name.lower() == "save_mode" or name.lower() == "__save_mode__":
                save_mode = SaveMode[value.lower()]
            if name.lower() in [
                "__create_hive_table__",
                "__create_output_table__",
            ]:
                create_output_table = value in [True, "true", "TRUE", "True", 1, "1"]

        if static_partition_name is not None:
            if static_partition_value is None or str(static_partition_value).strip() == "":
                raise SqlProcessorException(
                    f"partition value should exist and should not be empty, found {static_partition_value}"
                )
            target_table = Table(
                target_table_name, partitions=[Partition(field=static_partition_name, value=static_partition_value)]
            )
            self.collect_report(message=f"save with static partition: {static_partition_name}={static_partition_value}")
        else:
            target_table = Table(target_table_name)
            if not dry_run and backend.table_exists(target_table):
                backend.refresh_table_partitions(target_table)
            self.collect_report(message="save with dynamic partitions")

        if dry_run:
            if static_partition_name:
                if backend.is_spark_backend:
                    from pyspark.sql.functions import lit

                    table = table.with_column(static_partition_name, lit(static_partition_value))
                else:
                    table = table.with_column(static_partition_name, backend.sql_expr.for_value(static_partition_value))
            backend.create_temp_table(table, temp_table_name + "_output")
            self.collect_report(message="will not save data to data warehouse, since we are in dry run mode")
            return

        target_table_exists = backend.table_exists(target_table)
        if not target_table_exists and not create_output_table:
            message = f"target table {target_table.table_name} not exists"
            self.collect_report(message=message)
            raise Exception(message)

        backend.save_table(source_table, target_table, save_mode, create_target_table=create_output_table)

    def _write_for_log_step(self, df: BackendTable):
        log_data = df.limit(20).collect()
        if len(log_data) == 0:
            logger.info(f"log for [{self.target_config.name}]: no data to show")
            self.collect_report(message="no data to show")
        elif len(log_data) == 1:
            logger.info(f"log for [{self.target_config.name}]: {str(log_data[0])}")
            self.collect_report(message=f"{str(log_data[0])}")
        else:
            logger.info(f"log for [{self.target_config.name}]: ")
            df.show(20)
            self.collect_report(message=f"{str(log_data[0])}")

    def _write_for_check_step(self, df: BackendTable, context: ProcessorContext):
        if self.target_config.is_target_name_a_func():
            if not self.func_runner.run_func(self.target_config.name, context.vars_context):
                message = (
                    f"check failed! check function returned False. check={self.target_config.name}, vars={context.vars}"
                )
                self.collect_report(message=message)
                raise SqlProcessorException(message)
            else:
                return

        check_data = df.limit(100).collect()
        if not check_data:
            message = (
                "Data for check must contains at least one row. Please check your sql. "
                f"check={self.target_config.name}, check_data(limit 100)={check_data}, check_data_count={df.count()}"
            )
            self.collect_report(message=message)
            raise SqlProcessorException(message)
        for check_item in check_data:
            check_data_as_dict = check_item.as_dict()
            if "actual" not in check_data_as_dict or "expected" not in check_data_as_dict:
                message = (
                    "Data for check must contains expected and actual data. Please check your sql. "
                    f"check={self.target_config.name}, check_data(limit 100)={check_data}"
                )
                self.collect_report(message=message)
                raise SqlProcessorException(message)
            if check_data_as_dict["actual"] != check_data_as_dict["expected"]:
                message = (
                    f"check [{self.target_config.name}] failed! actual={check_data_as_dict['actual']},"
                    f" expected={check_data_as_dict['expected']}, check_data(limit 100)={check_data}"
                )
                logger.error(message)
                self.collect_report(message=message)
                raise SqlProcessorException(message)
        logger.info(f"check [{self.target_config.name}] passed! check_data(limit 100)={check_data}")
        self.collect_report(message=f"check_data(limit 100)={check_data}")

    def collect_report(self, status=None, message=None):
        self.reporter_collector.collect_report(self, status=status, message=message)


class StepFactory:
    def __init__(self, reporter: ReportCollector, func_runner: FuncRunner):
        self.reporter = reporter
        self.func_runner = func_runner

    def create_from_sql(self, sql: str, includes: Dict[str, str] = None) -> List[Step]:
        includes = includes or {}
        resolved_sql = self._resolve_include(sql, includes)
        lines = resolved_sql.split("\n")

        index = 0
        sql_parts = []
        step_list = []
        step = Step(f"step-{len(step_list) + 1}", self.reporter, self.func_runner)
        while index < len(lines):
            line = lines[index].replace(";", "")
            line_stripped = line.strip()
            if re.compile(StepConfig.STEP_CONFIG_PATTERN, flags=re.IGNORECASE).match(line_stripped):
                if len(sql_parts) > 0:
                    step.select_sql = "\n".join(sql_parts)
                if step.target_config is not None:
                    step_list.append(step)
                step = Step(f"step-{len(step_list) + 1}", self.reporter, self.func_runner)
                sql_parts = []
                target_config = StepConfig.from_config_line(line_stripped, index + 1)
                step.target_config = target_config
                if index == len(lines) - 1:
                    step_list.append(step)
            elif index == len(lines) - 1:
                if "" != line_stripped:
                    sql_parts.append(line)
                if len(sql_parts) > 0:
                    step.select_sql = "\n".join(sql_parts)
                step_list.append(step)
            elif "" != line_stripped:
                sql_parts.append(line)
            index += 1
        return step_list

    def _resolve_include(self, sql, includes: Dict[str, str]) -> str:
        include_sql_pattern = r"^--\s*include\s*=\s*(.*\.sql)\s*$"
        include_py_pattern = r"^--\s*include\s*=\s*(.*)\.(\w+|\*)$"
        lines = sql.split("\n")
        resoloved_sqls = []
        for _, line in enumerate(lines):
            line = line.replace(";", "")
            line_stripped = line.strip()
            if re.match(include_sql_pattern, line_stripped, flags=re.IGNORECASE):
                matches = re.match(include_sql_pattern, line_stripped, flags=re.IGNORECASE)
                if len(matches.groups()) != 1:
                    raise SqlProcessorException(
                        "parse include config failed. must provide complete module name and the sql variable name."
                        f"bug got config line {line_stripped}"
                    )
                file = matches.group(1)
                if file in includes:
                    resoloved_sqls.append(includes[file])
                    continue

                try:
                    import importlib

                    func_mod = importlib.import_module("common.file_reader")
                    read_file_func = func_mod.read_file
                    resoloved_sqls.append(read_file_func(file))
                except ModuleNotFoundError:
                    logger.info("failed to import common.file_reader, will try default file reader")
                    resoloved_sqls.append(SqlSnippetsReader.read_file(file))
            elif re.match(include_py_pattern, line_stripped, flags=re.IGNORECASE):
                matches = re.match(include_py_pattern, line_stripped, flags=re.IGNORECASE)
                if len(matches.groups()) != 2:
                    raise SqlProcessorException(
                        "parse include config failed. must provide complete module name and the sql variable name."
                        f"bug got config line {line_stripped}"
                    )
                module = matches.group(1)
                sql_name = matches.group(2)
                import importlib

                snippet_mod = importlib.import_module(module)
                resoloved_sqls.append(getattr(snippet_mod, sql_name))
            else:
                resoloved_sqls.append(line)
        resolved_sql = "\n".join(resoloved_sqls)
        if self._need_resolve(resolved_sql, include_sql_pattern, include_py_pattern):
            return self._resolve_include(resolved_sql, includes)
        return resolved_sql

    def _need_resolve(self, resolved_sql, include_sql_pattern, include_py_pattern):
        lines = resolved_sql.split("\n")
        return any(
            (
                self._match_sql_or_py_pattern(include_py_pattern, include_sql_pattern, line.replace(";", "").strip())
                for line in lines
            )
        )

    def _match_sql_or_py_pattern(self, include_py_pattern, include_sql_pattern, line):
        return re.match(include_sql_pattern, line, flags=re.IGNORECASE) or re.match(
            include_py_pattern, line, flags=re.IGNORECASE
        )


class SqlSnippetsReader:
    @staticmethod
    def read_file(file_name: str, base_path: str = None) -> str:
        possible_paths = [file_name]
        if base_path is not None:
            possible_paths.append(path.join(base_path, file_name))

        for p in possible_paths:
            if not path.exists(p):
                logger.info(f"file not found, tried: {p}")
                continue
            logger.info(f"read file at path: {p}")
            with open(p) as f:
                return f.read()

        raise FileNotFoundError(f"file not found: tried_paths={possible_paths}")



from __future__ import annotations

import re
import time
import unittest
from typing import TYPE_CHECKING

from easy_sql import base_test
from easy_sql.base_test import (
    TEST_BQ_URL,
    TEST_CH_URL,
    TEST_PG_URL,
    bigquery_sql_expr,
    date,
    dt,
)
from easy_sql.sql_processor.backend import Partition, SaveMode, TableMeta
from easy_sql.sql_processor.backend.rdb import RdbBackend, RdbRow, TimeLog, _exec_sql

if TYPE_CHECKING:
    from sqlalchemy.engine.reflection import Inspector


class RdbTest(unittest.TestCase):
    def test_log_time(self):
        with TimeLog("start", "end({time_took:.3f})"):
            time.sleep(0.1)

    def test_clean_pg_temp_schema(self):
        pg = RdbBackend(TEST_PG_URL)
        from sqlalchemy import inspect

        insp: Inspector = inspect(pg.engine)
        temp_schemas = insp.get_schema_names()
        for schema in temp_schemas:
            if schema.startswith("sp_temp_"):
                _exec_sql(pg.conn, f"drop schema if exists {schema} cascade")

    def test_clean_ch_temp_schema(self):
        backend = RdbBackend(TEST_CH_URL)
        from sqlalchemy import inspect

        insp: Inspector = inspect(backend.engine)
        temp_schemas = insp.get_schema_names()
        for schema in temp_schemas:
            if schema.startswith("sp_temp_"):
                _exec_sql(backend.conn, f"drop database if exists {schema}")

    def test_clean_bq_temp_schema(self):
        if not base_test.should_run_integration_test("bq"):
            return
        import os

        backend = RdbBackend(
            TEST_BQ_URL,
            credentials=f"{os.environ.get('HOME', '/tmp')}/.bigquery/credential-prod.json",
            sql_expr=bigquery_sql_expr,
        )
        from sqlalchemy import inspect

        insp: Inspector = inspect(backend.engine)
        temp_schemas = insp.get_schema_names()
        for schema in temp_schemas:
            if schema.startswith("sp_temp_"):
                _exec_sql(backend.conn, f"drop schema if exists {schema} cascade")

    def test_pg_table(self):
        backend = RdbBackend(TEST_PG_URL)
        _exec_sql(backend.conn, "drop schema if exists t cascade")
        _exec_sql(backend.conn, "create schema t")
        _exec_sql(backend.conn, "drop schema if exists t1 cascade")
        _exec_sql(backend.conn, "create schema t1")
        _exec_sql(backend.conn, "create table t.test(id int, val varchar(100))")
        _exec_sql(backend.conn, "insert into t.test values(1, '1'), (2, '2'), (3, '3')")
        self.run_test_table(backend)

    def test_ch_table(self):
        backend = RdbBackend(TEST_CH_URL)
        _exec_sql(backend.conn, "drop database if exists t")
        _exec_sql(backend.conn, "create database t")
        _exec_sql(backend.conn, "drop database if exists t1")
        _exec_sql(backend.conn, "create database t1")
        _exec_sql(backend.conn, "create table t.test(id int, val String) engine=MergeTree order by tuple()")
        _exec_sql(backend.conn, "insert into t.test values(1, '1'), (2, '2'), (3, '3')")
        self.run_test_table(backend)

    def test_bq_table(self):
        if not base_test.should_run_integration_test("bq"):
            return
        import os

        backend = RdbBackend(
            TEST_BQ_URL,
            credentials=f"{os.environ.get('HOME', '/tmp')}/.bigquery/credential-test.json",
            sql_expr=bigquery_sql_expr,
        )
        try:
            _exec_sql(backend.conn, "drop schema if exists t cascade")
        except Exception as e:
            # BigQuery will throw an exception when deleting a nonexistent dataset even if using [IF EXISTS]
            import re

            if not re.match(
                r"[\s\S]*Permission bigquery.datasets.delete denied on dataset[\s\S]*(or it may not exist)[\s\S]*",
                str(e.args[0]),
            ):
                raise e

        try:
            _exec_sql(backend.conn, "drop schema if exists t1 cascade")
        except Exception as e:
            import re

            if not re.match(
                r"[\s\S]*Permission bigquery.datasets.delete denied on dataset[\s\S]*(or it may not exist)[\s\S]*",
                str(e.args[0]),
            ):
                raise e

        _exec_sql(backend.conn, "create schema t")
        _exec_sql(backend.conn, "create schema t1")
        _exec_sql(backend.conn, "drop table if exists t.test")
        _exec_sql(backend.conn, "create table if not exists t.test(id int, val string)")
        _exec_sql(backend.conn, "insert into t.test values(1, '1'), (2, '2'), (3, '3')")

        from datetime import timedelta, timezone

        tz = timezone(timedelta(hours=0))
        self.run_test_table(backend, timezone=tz)

    def test_pg_backend(self):
        backend = RdbBackend(TEST_PG_URL)
        _exec_sql(backend.conn, "drop schema if exists t cascade")
        _exec_sql(backend.conn, "create schema t")
        _exec_sql(backend.conn, "create table t.test(id int, val varchar(100))")
        _exec_sql(backend.conn, "insert into t.test values(1, '1'), (2, '2'), (3, '3')")
        _exec_sql(backend.conn, "create table t.dynamic_partition_test(id int, val varchar(100), a varchar(100))")
        _exec_sql(
            backend.conn,
            "insert into t.dynamic_partition_test values(1, '1', '2021-01-01'), "
            "(2, '2', '2021-01-02'),(3, '3', '2021-01-03')",
        )

        self.run_test_backend(backend)

    def test_ch_backend(self):
        backend = RdbBackend(TEST_CH_URL)
        _exec_sql(backend.conn, "drop database if exists t")
        _exec_sql(backend.conn, "create database t")
        _exec_sql(backend.conn, "create table t.test(id int, val String) engine=MergeTree order by tuple()")
        _exec_sql(backend.conn, "insert into t.test values(1, '1'), (2, '2'), (3, '3')")
        _exec_sql(
            backend.conn,
            "create table t.dynamic_partition_test(id int, val String, a DateTime) engine=MergeTree order by tuple()",
        )
        _exec_sql(
            backend.conn,
            "insert into t.dynamic_partition_test values(1, '1', '2021-01-01 00:00:00'), "
            "(2, '2', '2021-01-02 00:00:00'), (3, '3', '2021-01-03 00:00:00')",
        )
        self.run_test_backend(backend)

    def test_bq_backend(self):
        if not base_test.should_run_integration_test("bq"):
            return
        import os

        backend = RdbBackend(
            TEST_BQ_URL,
            credentials=f"{os.environ.get('HOME', '/tmp')}/.bigquery/credential-test.json",
            sql_expr=bigquery_sql_expr,
        )
        _exec_sql(backend.conn, "drop schema if exists t cascade")
        _exec_sql(backend.conn, "create schema if not exists t")
        _exec_sql(backend.conn, "create table if not exists t.test(id int, val string)")
        _exec_sql(backend.conn, "insert into t.test values(1, '1'), (2, '2'), (3, '3')")
        _exec_sql(backend.conn, "create table if not exists t.dynamic_partition_test(id int, val string, a date)")
        _exec_sql(
            backend.conn,
            "insert into t.dynamic_partition_test values(1, '1', '2021-01-01'), (2, '2', '2021-01-02'), (3, '3',"
            " '2021-01-03')",
        )
        self.run_test_backend(backend)

    def run_test_table(self, backend: RdbBackend, timezone=None):
        table = backend.exec_sql("select * from t.test")
        self.assertFalse(table.is_empty())

        table = table.limit(0)
        self.assertTrue(table.is_empty())

        backend.exec_native_sql(backend.sql_dialect.rename_table_sql("t.test", "t.test1"))
        table = backend.exec_sql("select * from t.test1 order by id").with_column("a", backend.sql_expr.for_value(1))
        self.assertFalse(table.is_empty())
        self.assertTrue(table.count(), 3)
        self.assertListEqual(
            table.collect(),
            [
                RdbRow(["id", "val", "a"], (1, "1", 1)),
                RdbRow(["id", "val", "a"], (2, "2", 1)),
                RdbRow(["id", "val", "a"], (3, "3", 1)),
            ],
        )
        self.assertEqual(table.first(), RdbRow(["id", "val", "a"], (1, "1", 1)))
        self.assertListEqual(table.field_names(), ["id", "val", "a"])
        self.assertEqual(table.limit(2).count(), 2)

        backend.exec_native_sql(backend.sql_dialect.create_table_sql("t.test2", "select * from t.test1"))
        table = backend.exec_sql("select * from t.test2 order by id").with_column("a", backend.sql_expr.for_value("1"))
        self.assertEqual(table.first(), RdbRow(["id", "val", "a"], (1, "1", "1")))

        table = backend.exec_sql("select * from t.test2 order by id").with_column("a", backend.sql_expr.for_value(1.1))
        self.assertEqual(table.first(), RdbRow(["id", "val", "a"], (1, "1", 1.1)))

        table = backend.exec_sql("select * from t.test2 order by id").with_column(
            "a", backend.sql_expr.for_value(dt("2020-01-01 11:11:11"))
        )
        self.assertEqual(
            table.first(),
            RdbRow(["id", "val", "a"], (1, "1", base_test.dt_zone("2020-01-01 11:11:11", timezone=timezone))),
        )

        table.show(1)

        backend.exec_native_sql(backend.sql_dialect.create_view_sql("t.test2_view", "select * from t.test2"))
        backend.exec_native_sql(backend.sql_dialect.rename_view_sql("t.test2_view", "t.test2_view_new"))
        table = backend.exec_sql("select * from t.test2_view_new order by id")
        self.assertEqual(table.first(), RdbRow(["id", "val"], (1, "1")))

        backend.exec_native_sql(backend.sql_dialect.rename_table_db_sql("t.test2", "t1"))
        table = backend.exec_sql("select * from t1.test2 order by id")
        self.assertEqual(table.first(), RdbRow(["id", "val"], (1, "1")))

    def run_test_backend(self, backend: RdbBackend):
        backend.create_empty_table()  # should not raise exception

        backend.create_temp_table(backend.exec_sql("select * from t.test order by id limit 2"), "test_limit")
        self.assertListEqual(
            backend.exec_sql(f"select * from {backend.temp_schema}.test_limit order by id").collect(),
            [RdbRow(["id", "val"], (1, "1")), RdbRow(["id", "val"], (2, "2"))],
        )

        self.assertTrue(backend.table_exists(TableMeta("t.test")))
        self.assertFalse(backend.table_exists(TableMeta("t.test_xx")))
        self.assertTrue(backend.table_exists(TableMeta("test_limit")))

        # save without transformation or partitions, should create
        backend.create_cache_table(backend.exec_sql("select * from t.test"), "test")
        backend.save_table(TableMeta("test"), TableMeta("t.xx0"), SaveMode.overwrite, True)
        self.assertListEqual(
            backend.exec_sql("select * from t.xx0 order by id").collect(),
            [RdbRow(["id", "val"], (1, "1")), RdbRow(["id", "val"], (2, "2")), RdbRow(["id", "val"], (3, "3"))],
        )

        self.assertRaisesRegex(
            Exception,
            re.compile(r".* cannot save table.*"),
            lambda: backend.save_table(TableMeta("test_limit"), TableMeta("t.xx"), SaveMode.overwrite, False),
        )

        # first save without partitions, should create
        backend.save_table(TableMeta("test_limit"), TableMeta("t.xx"), SaveMode.overwrite, True)
        self.assertListEqual(
            backend.exec_sql("select * from t.xx order by id").collect(),
            [RdbRow(["id", "val"], (1, "1")), RdbRow(["id", "val"], (2, "2"))],
        )

        # second save without partitions, should overwrite
        backend.exec_native_sql(backend.sql_dialect.drop_view_sql("test_limit"))
        backend.create_cache_table(backend.exec_sql("select * from t.test order by id limit 2"), "test_limit")
        backend.save_table(TableMeta("test_limit"), TableMeta("t.xx"), SaveMode.overwrite, True)
        self.assertListEqual(
            backend.exec_sql("select * from t.xx order by id").collect(),
            [RdbRow(["id", "val"], (1, "1")), RdbRow(["id", "val"], (2, "2"))],
        )

        # third save without partitions, should create extra row
        backend.exec_native_sql(backend.sql_dialect.drop_view_sql("test_limit"))
        backend.create_cache_table(backend.exec_sql("select * from t.test order by id limit 3"), "test_limit")
        backend.save_table(TableMeta("test_limit"), TableMeta("t.xx"), SaveMode.overwrite, True)
        self.assertListEqual(
            backend.exec_sql("select * from t.xx order by id").collect(),
            [RdbRow(["id", "val"], (1, "1")), RdbRow(["id", "val"], (2, "2")), RdbRow(["id", "val"], (3, "3"))],
        )

        if backend.is_bq:
            mock_dt_1, mock_dt_2, mock_dt_3 = date("2021-01-01"), date("2021-01-02"), date("2021-01-03")
        elif backend.is_pg:
            mock_dt_1, mock_dt_2, mock_dt_3 = "2021-01-01", "2021-01-02", "2021-01-03"
        else:
            mock_dt_1, mock_dt_2, mock_dt_3 = (
                dt("2021-01-01 00:00:00"),
                dt("2021-01-02 00:00:00"),
                dt("2021-01-03 00:00:00"),
            )

        backend.exec_native_sql(backend.sql_dialect.drop_view_sql("test_limit"))
        backend.create_cache_table(backend.exec_sql("select id from t.test limit 2"), "test_limit")
        self.assertRaisesRegex(
            Exception,
            re.compile(r"source_cols does not contain target_cols.*"),
            lambda: backend.save_table(
                TableMeta("test_limit"),
                TableMeta("t.xx", partitions=[Partition("a", mock_dt_1)]),
                SaveMode.overwrite,
                True,
            ),
        )

        _exec_sql(backend.conn, "drop table if exists t.xx")

        # first save with partitions, should create
        backend.exec_native_sql(backend.sql_dialect.drop_view_sql("test_limit"))
        backend.create_cache_table(backend.exec_sql("select * from t.test order by id limit 2"), "test_limit")
        backend.save_table(
            TableMeta("test_limit"), TableMeta("t.xx", partitions=[Partition("a", mock_dt_1)]), SaveMode.overwrite, True
        )
        self.assertListEqual(
            backend.exec_sql("select * from t.xx order by id").collect(),
            [RdbRow(["id", "val", "a"], (1, "1", mock_dt_1)), RdbRow(["id", "val", "a"], (2, "2", mock_dt_1))],
        )

        # second save with partitions, should overwrite
        backend.exec_native_sql(backend.sql_dialect.drop_view_sql("test_limit"))
        backend.create_cache_table(backend.exec_sql("select * from t.test order by id limit 2"), "test_limit")
        backend.save_table(
            TableMeta("test_limit"), TableMeta("t.xx", partitions=[Partition("a", mock_dt_2)]), SaveMode.overwrite, True
        )
        self.assertListEqual(
            backend.exec_sql("select * from t.xx order by a, id").collect(),
            [
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_1)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_1)),
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_2)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_2)),
            ],
        )

        # third save with partitions, should overwrite
        backend.exec_native_sql(backend.sql_dialect.drop_view_sql("test_limit"))
        backend.create_cache_table(backend.exec_sql("select * from t.test order by id limit 2"), "test_limit")
        backend.save_table(
            TableMeta("test_limit"),
            TableMeta("t.xx", partitions=[Partition("a", mock_dt_2)]),
            SaveMode.overwrite,
            False,
        )
        self.assertListEqual(
            backend.exec_sql("select * from t.xx order by a, id").collect(),
            [
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_1)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_1)),
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_2)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_2)),
            ],
        )

        # fourth save with partitions, should append
        backend.exec_native_sql(backend.sql_dialect.drop_view_sql("test_limit"))
        backend.create_cache_table(backend.exec_sql("select * from t.test order by id limit 2"), "test_limit")
        backend.save_table(
            TableMeta("test_limit"), TableMeta("t.xx", partitions=[Partition("a", mock_dt_2)]), SaveMode.append, True
        )
        self.assertListEqual(
            backend.exec_sql("select * from t.xx order by a, id").collect(),
            [
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_1)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_1)),
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_2)),
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_2)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_2)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_2)),
            ],
        )

        # first save with dynamic partitions, create
        _exec_sql(backend.conn, "drop table if exists t.xx1")
        backend.create_cache_table(
            backend.exec_sql("select * from t.dynamic_partition_test order by id limit 3"),
            "dynamic_partition_test_limit",
        )
        backend.save_table(
            TableMeta("dynamic_partition_test_limit"),
            TableMeta("t.xx1", partitions=[Partition("a", None)]),
            SaveMode.overwrite,
            True,
        )
        self.assertListEqual(
            backend.exec_sql("select * from t.xx1 order by a, id").collect(),
            [
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_1)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_2)),
                RdbRow(["id", "val", "a"], (3, "3", mock_dt_3)),
            ],
        )

        # second save with dynamic partitions, overwrite
        backend.exec_native_sql(backend.sql_dialect.drop_view_sql("dynamic_partition_test_limit"))
        backend.create_cache_table(
            backend.exec_sql("select * from t.dynamic_partition_test order by id limit 3"),
            "dynamic_partition_test_limit",
        )
        backend.save_table(
            TableMeta("dynamic_partition_test_limit"),
            TableMeta("t.xx1", partitions=[Partition("a", None)]),
            SaveMode.overwrite,
            False,
        )
        self.assertListEqual(
            backend.exec_sql("select * from t.xx1 order by a, id").collect(),
            [
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_1)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_2)),
                RdbRow(["id", "val", "a"], (3, "3", mock_dt_3)),
            ],
        )

        # third save with dynamic partitions, append
        backend.exec_native_sql(backend.sql_dialect.drop_view_sql("dynamic_partition_test_limit"))
        backend.create_cache_table(
            backend.exec_sql("select * from t.dynamic_partition_test order by id limit 3"),
            "dynamic_partition_test_limit",
        )
        backend.save_table(
            TableMeta("dynamic_partition_test_limit"),
            TableMeta("t.xx1", partitions=[Partition("a", None)]),
            SaveMode.append,
            False,
        )
        self.assertListEqual(
            backend.exec_sql("select * from t.xx1 order by a, id").collect(),
            [
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_1)),
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_1)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_2)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_2)),
                RdbRow(["id", "val", "a"], (3, "3", mock_dt_3)),
                RdbRow(["id", "val", "a"], (3, "3", mock_dt_3)),
            ],
        )

        # fourth save with dynamic partitions, overwrite
        backend.exec_native_sql(backend.sql_dialect.drop_view_sql("dynamic_partition_test_limit"))
        backend.create_cache_table(
            backend.exec_sql("select * from t.dynamic_partition_test order by id limit 3"),
            "dynamic_partition_test_limit",
        )
        backend.save_table(
            TableMeta("dynamic_partition_test_limit"),
            TableMeta("t.xx1", partitions=[Partition("a", None)]),
            SaveMode.overwrite,
            False,
        )
        self.assertListEqual(
            backend.exec_sql("select * from t.xx1 order by a, id").collect(),
            [
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_1)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_2)),
                RdbRow(["id", "val", "a"], (3, "3", mock_dt_3)),
            ],
        )

        # first save with non-existing target_table, create
        backend.exec_native_sql(backend.sql_dialect.drop_view_sql("dynamic_partition_test_limit"))
        backend.create_cache_table(
            backend.exec_sql("select * from t.dynamic_partition_test order by id limit 3"),
            "dynamic_partition_test_limit",
        )
        backend.save_table(
            TableMeta("dynamic_partition_test_limit"),
            TableMeta("t_created_db.created_table", partitions=[Partition("a", None)]),
            SaveMode.overwrite,
            True,
        )
        self.assertListEqual(
            backend.exec_sql("select * from t_created_db.created_table order by a, id").collect(),
            [
                RdbRow(["id", "val", "a"], (1, "1", mock_dt_1)),
                RdbRow(["id", "val", "a"], (2, "2", mock_dt_2)),
                RdbRow(["id", "val", "a"], (3, "3", mock_dt_3)),
            ],
        )

        pre_temp_db = backend.temp_schema
        backend.reset()
        cur_temp_db = backend.temp_schema
        self.assertNotEqual(pre_temp_db, cur_temp_db)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import json
import os
from typing import TYPE_CHECKING, Dict, List, Union

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession
    from pyspark.sql.types import Row
    from easy_sql.sql_processor.context import ProcessorContext

from .backend import SparkBackend
from .common import SqlProcessorAssertionError, _exec_sql, is_int_type
from .funcs_common import AlertFunc, ColumnFuncs
from .funcs_common import PartitionFuncs as PartitionFuncsBase
from .funcs_common import TableFuncs

__all__ = [
    "PartitionFuncs",
    "ColumnFuncs",
    "CacheFuncs",
    "ParallelismFuncs",
    "IOFuncs",
    "AlertFunc",
    "TableFuncs",
    "ModelFuncs",
]

from ..logger import logger


class ParallelismFuncs:
    def __init__(self, spark: SparkSession):
        self.spark = spark

    def repartition(self, table: str, partitions: str):
        try:
            partitions = int(partitions)
        except ValueError:
            raise Exception(f"partitions must be an int when repartition a table, got `{partitions}`")
        _exec_sql(self.spark, f"select * from {table}").repartition(partitions).createOrReplaceTempView(table)

    def repartition_by_column(self, table: str, partitions: str):
        _exec_sql(self.spark, f"select * from {table}").repartition(partitions).createOrReplaceTempView(table)

    def coalesce(self, table: str, partitions: str):
        try:
            partitions = int(partitions)
        except ValueError:
            raise Exception(f"partitions must be an int when coalesce a table, got `{partitions}`")
        _exec_sql(self.spark, f"select * from {table}").coalesce(partitions).createOrReplaceTempView(table)

    def set_shuffle_partitions(self, partitions: str):
        self.spark.conf.set("spark.sql.adaptive.enabled", "false")
        self.spark.conf.set("spark.sql.shuffle.partitions", partitions)


class IOFuncs:
    def __init__(self, spark: SparkSession):
        self.spark = spark

    def write_csv(self, table: str, output_file: str):
        _exec_sql(self.spark, f"select * from {table}").repartition(1).write.mode("overwrite").csv(
            output_file, header=True
        )

    def rename_csv_output(self, spark_output_path: str, to_file: str):
        import re
        import subprocess

        assert to_file.startswith("/"), "to_file must be a full path starts with /, found: " + to_file
        is_local_file = spark_output_path.startswith("file:///")
        if is_local_file:
            command_prefix = ""
            spark_output_path = spark_output_path[len("file://") :]
        else:
            command_prefix = "hdfs dfs -"
        result = (
            subprocess.check_output(f"{command_prefix}ls {spark_output_path}".split(" ")).decode("utf8").split("\n")
        )
        generated_files = []
        for line in result:
            if line.endswith(".csv"):
                file_path = re.sub(r".*[0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}", "", line).strip()
                # hdfs command will return full path, but system command will not, try fix to full path
                file_path = os.path.join(spark_output_path, file_path) if is_local_file else file_path
                generated_files.append(file_path)
        if not generated_files:
            raise Exception("no csv file found at path: " + spark_output_path)
        if len(generated_files) != 1:
            raise SqlProcessorAssertionError(
                f"found multiple csv file at path: {spark_output_path}, files={generated_files}"
            )
        subprocess.check_output(f"{command_prefix}mkdir -p {os.path.dirname(to_file)}".split(" "))
        try:
            subprocess.check_output(f"{command_prefix}rm {to_file}".split(" "))
        except Exception:
            logger.info(f"{to_file} not found, remove skipped")
        subprocess.check_output(f"{command_prefix}mv {generated_files[0]} {to_file}".split(" ")).decode("utf8").split(
            "\n"
        )

    def write_json_local(self, table: str, output_file: str):
        data: DataFrame = _exec_sql(self.spark, f"select * from {table}")
        data: List[Row] = data.collect()
        data: List[Dict] = [row.asDict() for row in data]
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, "w") as f:
            f.write(json.dumps(data, ensure_ascii=False, indent=4, sort_keys=False))

    def update_json_local(self, context, vars: str, list_vars: str, json_attr: str, output_file: str):
        context: ProcessorContext = context
        vars_value = {
            var.strip(): context.vars_context.vars.get(var.strip(), None) for var in vars.split(",") if var.strip()
        }
        list_vars_value = {
            var.strip(): context.vars_context.list_vars.get(var.strip(), None)
            for var in list_vars.split(",")
            if var.strip()
        }

        data = {}
        if os.path.exists(output_file):
            with open(output_file, "r") as f:
                data = json.loads(f.read())

        from easy_sql.utils.object_utils import get_attr

        data_current = get_attr(data, json_attr)
        data_current.update(vars_value)
        data_current.update(list_vars_value)

        with open(output_file, "w") as f:
            f.write(json.dumps(data, ensure_ascii=False, indent=4, sort_keys=False))


class ModelFuncs:
    def __init__(self, spark: SparkSession):
        self.spark = spark

    def model_predict(
        self, model_save_path: str, table_name: str, feature_cols: str, id_col: str, output_ref_cols: str
    ):
        from pyspark.ml import PipelineModel
        from pyspark.sql.functions import expr

        output_ref_cols = [col.strip() for col in output_ref_cols.split(",") if col.strip()]
        model = PipelineModel.load(model_save_path)
        data = _exec_sql(self.spark, f"select {feature_cols} from {table_name}")

        int_cols = [f.name for f in data.schema.fields if is_int_type(f.dataType.typeName())]
        for col in int_cols:
            data = data.withColumn(col, expr(f"cast({col} as double)"))

        predictions = model.transform(data)
        output = predictions.select(output_ref_cols + [id_col, "prediction"])
        output.createOrReplaceTempView(table_name)


class CacheFuncs:
    def __init__(self, spark: SparkSession):
        self.spark = spark

    def unpersist(self, table_name: str):
        self.spark.catalog.uncacheTable(table_name)


class PartitionFuncs(PartitionFuncsBase):
    def __init__(self, backend: Union[SparkSession, SparkBackend]):
        backend = backend if isinstance(backend, SparkBackend) else SparkBackend(backend)
        super().__init__(backend)

    def _get_partition_values(self, table_name):
        backend: SparkBackend = self.backend
        partitions = _exec_sql(backend.spark, f"show partitions {table_name}").collect()
        partition_values = [p[0][p[0].index("=") + 1 :] for p in partitions]
        return partition_values

    def get_partition_cols(self, table_name: str) -> List[str]:
        backend: SparkBackend = self.backend
        all_cols = _exec_sql(backend.spark, f"desc table {table_name}").collect()
        pt_cols = []
        pt_col_start = False
        for col in all_cols:
            if pt_col_start:
                pt_cols.append(col[0].strip())
            if col[0].strip() == "# col_name":
                pt_col_start = True
        return pt_cols



import unittest
from decimal import Decimal
from typing import Dict

from pyspark.sql.types import IntegerType, StructField, StructType
from pyspark.sql.utils import ParseException

from easy_sql.base_test import LocalSpark, run_sql

from .sql_processor import FuncRunner, SqlProcessor, SqlProcessorException, VarsContext
from .sql_processor.step import StepFactory


class SqlProcessorTest(unittest.TestCase):
    def test_process_sql(self):
        sql = """
-- target=template.test_a
select
    *
from target
where
    type = 'a';
-- target=variables
select 1 as a
-- target=variables
select '${f2(1, 2)}' as b
-- target=log.b
select '${b}' as b
-- target=log.test_log
select 1 as t
-- target=check.test_check
select 1 as actual, 1 as expected
-- target=check.check(1, 2), if=f1(1, ${a})
-- target=template.test_b
select
    id as id,
    #{type} as type
from target
where
    type = #{type};
-- target=template.test_c
select
    id as id,
    #{type2} as type
from target
where
    type = #{type1};

-- target=broadcast.test_result
select * from (
    @{test_a()}
) -- comment
--
union all
select * from (
    @{test_b(type='b')}
)
union all
select * from (
    @{test_c(type1='c', type2='${c}')}
)
        """
        spark = LocalSpark.get()

        df = spark.createDataFrame([(1, "a"), (2, "b"), (3, "c"), (4, "d"), (5, "e")], ["id", "type"])
        df.createOrReplaceTempView("target")

        variables = {"c": "c"}
        funcs = {
            "check": lambda a, b: a and b,
            "f1": lambda a, b: a and b,
            "f2": lambda a, b: a + b,
        }
        self.assertListEqual(
            run_sql(sql, "test_result", funcs=funcs, variables=variables, spark=spark),
            [
                (1, "a"),
                (2, "b"),
                (3, "c"),
            ],
        )

    def test_list_var(self):
        sql = """
-- target=list_variables
select explode(array(1, 2, 3)) as a
-- target=list_variables
select explode(array('1', '2', '3')) as b
        """
        spark = LocalSpark.get()
        processor = SqlProcessor(spark, sql, [], {})
        processor.run()
        self.assertEqual(processor.context.vars_context.list_vars, {"a": [1, 2, 3], "b": ["1", "2", "3"]})

    def test_var_replace(self):
        sql = """
-- target=variables
select 1 as a
, 2 as aa
-- target=variables
select '${f1(${a}, 2)}, ${a}' as b
-- target=variables, if=bool()
select '1' as b
-- target=temp.result, if=f1(,)
select '${b}-${aa}-dont-run' as b
-- target=temp.result
select '${b}-${aa}' as b
        """
        self.assertEqual(run_sql(sql, "result", funcs={"f1": lambda a, b: a + b}), [("12, 1-2",)])

    def test_template_with_special_chars(self):
        # template引用时，参数值如果是字符串，则字符串中不能出现字符`()`，如果有此需求，可以使用变量支持(变量值内部可以有这些符号)
        sql = """
-- target=variables
select 'abc(),' as a,
        '1 as col1, 2 as col2, \\'col-3\\' as col3' as other_cols
-- target=template.templ
select #{a} as a, #{b} as b, #{c} as c, #{other_cols}
-- target=temp.result
@{templ(a='${a}', b=2, c='3', other_cols=${other_cols})}
        """
        self.assertEqual(run_sql(sql, "result", funcs={"c": lambda a, b: a + b}), [("abc(),", 2, "3", 1, 2, "col-3")])

    def test_template_replace_mix_var_replace(self):
        # template中可以有variable引用，解析时会先解析template，再替换variable
        sql = """
-- target=template.temp1
select #{a} as a, #{b} as b, #{c} as c, ${d} as d
-- target=cache.result
@{temp1(a=1, b=2,c=3)}
        """
        self.assertEqual(run_sql(sql, "result", variables={"d": "4"}), [(1, 2, 3, 4)])

        # 可以多次引用template
        sql = """
-- target=template.temp1
#{a} as #{an}, #{b} as #{bn}
-- target=cache.result
select @{temp1(a=1, an=a, b=2, bn=b)},
@{temp1(a=3, an=c, b=4, bn=d)}
        """
        self.assertEqual(run_sql(sql, "result", variables={"d": "4"}), [(1, 2, 3, 4)])

        # 可以在template中引用template, 但是不推荐这样用，会导致逻辑很复杂，理解不易，且容易导致循环引用
        sql = """
-- target=template.temp1
1 as a, #{b} as b
-- target=template.temp2
select @{temp1(b=${b})}
-- target=cache.result
@{temp2}
        """
        self.assertEqual(run_sql(sql, "result", variables={"b": "2"}), [(1, 2)])

        # template引用时，如果没有参数时可以省略括号
        sql = """
-- target=template.templ
select 1 as a
-- target=cache.result
@{templ} union all @{templ()}
        """
        self.assertEqual(run_sql(sql, "result"), [(1,), (1,)])

        # template引用时，参数可以换行
        sql = """
-- target=template.temp1
#{a} as a,
    #{b} as b, #{c} as c, ${d} as d, ${d1(1, ${d})} as e
-- target=cache.result
select
@{temp1(
a=1.1, b='2-2',
    c=3)}
            """
        self.assertEqual(
            run_sql(sql, "result", funcs={"d1": lambda a, b: a + b}, variables={"d": "4"}),
            [(Decimal("1.1"), "2-2", 3, 4, 14)],
        )

        # template引用时，参数值可以是变量
        sql = """
-- target=template.templ
select #{a} as a, #{b} as b, #{c} as c
-- target=temp.result
@{templ(a=${a}, b=2, c=${a})}
        """
        self.assertEqual(run_sql(sql, "result", variables={"a": "1"}), [(1, 2, 1)])

        # template引用时，参数值如果是变量，则变量中不能有函数
        sql = """
-- target=template.templ
select #{a} as a, #{b} as b, #{c} as c
-- target=temp.result
@{templ(a=${a}, b=2, c=${c(1, 2)})}
        """
        with self.assertRaises(expected_exception=ParseException):
            run_sql(sql, "result", funcs={"c": lambda a, b: a + b}, variables={"a": "1"})

        # 可以调用一个函数
        sql = """
-- target=func.t(a, ${b})
-- target=temp.result
select 1 as a
            """
        c = ""

        def func(a, b):
            nonlocal c
            c = a + b

        self.assertEqual(run_sql(sql, "result", funcs={"t": func}, variables={"b": "2"}), [(1,)])
        self.assertEqual(c, "a2")

        # dryrun运行时，只会打印日志，不会真的写入hive库
        sql = """
-- target=temp.result
select 1 as a, 2 as b
-- target=hive.db.result
select 1 as a, 2 as b
            """
        self.assertEqual(run_sql(sql, "result", dry_run=True), [(1, 2)])

        # skip_all生效时，剩下的步骤均不会执行
        sql = """
-- target=variables
select '${f(0)}' as __skip_all__
-- target=temp.result
select 1 as a, 2 as b
-- target=variables
select '${f(1)}' as __skip_all__
-- target=temp.result
select 11 as a, 22 as b
            """
        self.assertEqual(run_sql(sql, "result", funcs={"f": lambda x: x == "1"}), [(1, 2)])

        # 可以给函数传入元变量，如 __step__
        sql = """
-- target=func.ensure_partition_exists(${__step__}, a, b, c, 1)
-- target=temp.result
select 1, 2 as dt
-- target=func.ensure_no_null_data_in_table(${__step__}, result, dt=${d})
"""
        self.assertEqual(run_sql(sql, "result", variables={"d": "1"}), [(1, 2)])

    def test_should_be_able_to_parse_target_when_target_is_at_last(self):
        sql = """
-- target=temp.result
select 1 as a
-- target=func.t(a, ${b})
-- target=func.t(a, ${b})"""
        c = 0

        def func(a, b):
            nonlocal c
            c += 1

        self.assertEqual(run_sql(sql, "result", funcs={"t": func}, variables={"b": "2"}), [(1,)])
        self.assertEqual(c, 2)

    def test_should_process_exception_when_exception_handler_not_null(self):
        sql = """
-- target=variables
select 'old_a' as var_a

-- case1: exception_handler 可以带参数，使用 `{var_name}` 来引用变量，变量值取 handler 被执行时的取值
-- target=variables
select "exception_handler1({__step__}, {var_a}, b)" as __exception_handler__
-- target=variables
select 'current_a' as var_a
-- target=temp.res1
select ${not_existed1}

-- case2: exception_handler 可以不带参数
-- target=variables
select 'exception_handler2()' as __exception_handler__
-- target=temp.res2
select ${not_existed2}

-- case3: 可以针对接下来的步骤取消 exception_handler 设定，将值置为 null 即可
-- target=variables
select null as __exception_handler__
-- target=temp.res3
select ${not_existed3}
"""
        from .sql_processor import SqlProcessor, StepReport, StepStatus
        from .sql_processor.step import Step

        def exception_handler1(step: Step, a, b):
            return lambda e: step.collect_report(status=StepStatus.FAILED, message=f"error: {a}, {b}, {e}")

        def exception_handler2():
            return lambda e: print(f"error: {e}")

        spark = LocalSpark.get()
        processor = SqlProcessor(spark, sql)
        processor.register_funcs({"exception_handler1": exception_handler1, "exception_handler2": exception_handler2})

        self.assertRaisesRegex(Exception, ".*not_existed3.*", processor.run)

        reports: Dict[str, StepReport] = processor.reporter.step_reports
        self.assertIn("error: current_a, b, ", reports["step-4"].report_as_text(1))
        self.assertEqual(StepStatus.FAILED, reports["step-4"].status)
        self.assertEqual(StepStatus.FAILED, reports["step-6"].status)

    def test_should_raise_exception_when_check_failed(self):
        check_fail_sqls = [
            "-- target=check.test_check\nselect 1 as actual, 0 as expected",
            "-- target=check.test_check\nselect 1 as actual, 1 as expected1",
            "-- target=check.test_check\nselect 1 as actual, 0 as expected where 1=0",
        ]
        for sql in check_fail_sqls:
            processor = SqlProcessor(LocalSpark.get(), sql)
            self.assertRaises(SqlProcessorException, processor.run)

    def test_should_not_fail_and_log_no_data_when_no_data_found_in_log_target(self):
        processor = SqlProcessor(LocalSpark.get(), "-- target=log.no_data\nselect 1 as actual, 0 as expected where 1=0")
        processor.run()

    def test_should_raise_error_when_no_table_and_no_create_output_table(self):
        processor = SqlProcessor(LocalSpark.get(), "-- target=output.t.some_table\nselect 1 as a, 0 as b")
        self.assertRaises(Exception, lambda: processor.run())

    def test_should_raise_error_when_output_table_name_does_not_contain_dbname(self):
        processor = SqlProcessor(LocalSpark.get(), "-- target=output.some_table\nselect 1 as actual, 0 as expected")
        self.assertRaises(Exception, lambda: processor.run())

    def test_should_do_action_step_ok(self):
        processor = SqlProcessor(
            LocalSpark.get(),
            "-- target=variables\nselect true as __create_output_table__\n"
            "-- target=output.t.some_table\nselect 1 as a, 0 as b\n"
            "-- target=action.some_action\ndrop table t.some_table",
        )
        processor.run()
        from easy_sql.sql_processor.backend import TableMeta

        self.assertFalse(processor.backend.table_exists(TableMeta("t.some_table")))

    def test_should_add_static_partition_value_for_dryrun_as_well(self):
        sql = """
-- target=variables
select 20200101 as __partition__dt
-- target=output.t.result
select 1 as a, 2 as b
        """
        spark = LocalSpark.get()
        processor = SqlProcessor(spark, sql)
        processor.run(True)
        result = [t for t in processor.backend.temp_tables() if t.startswith("result_") and t.endswith("_output")]
        self.assertEqual(len(result), 1)
        self.assertEqual(spark.sql(f"select * from {result[0]}").collect(), [(1, 2, "20200101")])


class FuncRunnerTest(unittest.TestCase):
    def test_fun_func(self):
        def f1(a, b) -> bool:
            return a and b

        self.assertFalse(FuncRunner({"f1": f1}).run_func("f1(,)", VarsContext()))
        self.assertTrue(FuncRunner({"f1": f1}).run_func("f1(1,1)", VarsContext()))
        self.assertTrue(FuncRunner({"f1": f1}).run_func("f1(1,${a})", VarsContext({"a": 1})))
        self.assertFalse(FuncRunner({"f1": f1}).run_func("f1(1,${a})", VarsContext({"a": ""})))

    def test_ensure_dwd_partition_exists(self):
        spark = LocalSpark.get()
        spark.createDataFrame(
            [], StructType([StructField("id", IntegerType()), StructField("pt", IntegerType())])
        ).write.mode("overwrite").partitionBy("pt").saveAsTable("empty_table")
        spark.createDataFrame(
            [
                (
                    1,
                    None,
                    None,
                    20210101,
                ),
                (
                    1,
                    "1",
                    "1",
                    20210102,
                ),
            ],
            ["id", "fk1", "fk2", "pt"],
        ).write.mode("overwrite").partitionBy("pt").saveAsTable("data_table")
        spark.createDataFrame([(1, "1", 20210101,)], ["id", "fk", "pt"]).write.mode("overwrite").partitionBy(
            "pt"
        ).saveAsTable("data_table1")
        # failure case1: 被检测表是空的
        with self.assertRaises(Exception):  # noqa: B017
            processor = SqlProcessor(
                spark, "-- target=check.ensure_dwd_partition_exists(${__step__}, empty_table, 20210101)", [], {}
            )
            processor.run(dry_run=True)
        self.assertIn("partition 20210101 not exists: empty_table", processor.reporter.step_reports["step-1"].messages)

        # failure case2: 检测分区在被检测表中确实不存在
        with self.assertRaises(Exception):  # noqa: B017
            processor = SqlProcessor(
                spark, "-- target=check.ensure_dwd_partition_exists(${__step__}, data_table, 20210103)", [], {}
            )
            processor.run(dry_run=True)
        self.assertIn("partition 20210103 not exists: data_table", processor.reporter.step_reports["step-1"].messages)

        # happy case1: 检测分区存在
        processor = SqlProcessor(
            spark, "-- target=check.ensure_dwd_partition_exists(${__step__}, data_table, 20210101)", [], {}
        )
        processor.run(dry_run=True)

        # happy case2: 检测分区 < first_partition
        processor = SqlProcessor(
            spark, "-- target=check.ensure_dwd_partition_exists(${__step__}, data_table, 20201201)", [], {}
        )
        processor.run(dry_run=True)

        # failure case: 检测分区存在，但全部外键均为空
        with self.assertRaises(Exception):  # noqa: B017
            processor = SqlProcessor(
                spark,
                "-- target=check.ensure_dwd_partition_exists(${__step__}, data_table, 20210101, fk1, fk2)",
                [],
                {},
            )
            processor.run(dry_run=True)
        self.assertIn(
            "all fk cols are null in partition: table_name=data_table, partition=20210101",
            processor.reporter.step_reports["step-1"].messages,
        )

        # failure case: 检测分区 < first_partition，全部外键均为空
        with self.assertRaises(Exception):  # noqa: B017
            processor = SqlProcessor(
                spark,
                "-- target=check.ensure_dwd_partition_exists(${__step__}, data_table, 20201101, fk1, fk2)",
                [],
                {},
            )
            processor.run(dry_run=True)
        self.assertIn(
            "all fk cols are null in partition: table_name=data_table, partition=20210101",
            processor.reporter.step_reports["step-1"].messages,
        )

        # happy case: 检测分区存在，且存在外键不为空的数据
        processor = SqlProcessor(
            spark, "-- target=check.ensure_dwd_partition_exists(${__step__}, data_table, 20210102, fk1, fk2)", [], {}
        )
        processor.run(dry_run=True)

        # happy case: 检测分区 < first_partition，且存在外键不为空的数据
        processor = SqlProcessor(
            spark, "-- target=check.ensure_dwd_partition_exists(${__step__}, data_table1, 20201201, fk)", [], {}
        )
        processor.run(dry_run=True)

    def test_ensure_partition_or_first_partition_exists(self):
        spark = LocalSpark.get()
        spark.createDataFrame(
            [], StructType([StructField("id", IntegerType()), StructField("pt", IntegerType())])
        ).write.mode("overwrite").partitionBy("pt").saveAsTable("empty_table")
        spark.createDataFrame([(1, 20210101,)], ["id", "pt"]).write.mode("overwrite").partitionBy(
            "pt"
        ).saveAsTable("data_table")
        # failure case1: 被检测表是空的
        with self.assertRaises(Exception):  # noqa: B017
            processor = SqlProcessor(
                spark,
                "-- target=check.ensure_partition_or_first_partition_exists(${__step__}, empty_table, 20210101)",
                [],
                {},
            )
            processor.run(dry_run=True)
        self.assertIn(
            "partition 20210101 not exists: ['empty_table']", processor.reporter.step_reports["step-1"].messages
        )

        # failure case2: 检测分区在被检测表中确实不存在
        with self.assertRaises(Exception):  # noqa: B017
            processor = SqlProcessor(
                spark,
                "-- target=check.ensure_partition_or_first_partition_exists(${__step__}, data_table, 20210102)",
                [],
                {},
            )
            processor.run(dry_run=True)
        self.assertIn(
            "partition 20210102 not exists: ['data_table']", processor.reporter.step_reports["step-1"].messages
        )

        # failure case3: 检测多张表，message 中应该只有检查失败的表
        with self.assertRaises(Exception):  # noqa: B017
            processor = SqlProcessor(
                spark,
                "-- target=check.ensure_partition_or_first_partition_exists(${__step__}, empty_table, data_table,"
                " 20210101)",
                [],
                {},
            )
            processor.run(dry_run=True)
        self.assertIn(
            "partition 20210101 not exists: ['empty_table']", processor.reporter.step_reports["step-1"].messages
        )

        # happy case1: 检测分区存在
        processor = SqlProcessor(
            spark,
            "-- target=check.ensure_partition_or_first_partition_exists(${__step__}, data_table, 20210101)",
            [],
            {},
        )
        processor.run(dry_run=True)

        # happy case2: 检测分区 < first_partition
        processor = SqlProcessor(
            spark,
            "-- target=check.ensure_partition_or_first_partition_exists(${__step__}, data_table, 20201201)",
            [],
            {},
        )
        processor.run(dry_run=True)


test_snippet = """
-- target=variables
select 1 as a
"""
test_snippet2 = """
-- target=variables
select 1 as b
"""


class StepFactoryTest(unittest.TestCase):
    def test_include_snippet_from_py(self):
        spark = LocalSpark.get()
        sql = """
-- can have comments here
-- include=easy_sql.sql_processor_test.test_snippet
-- target=temp.result1
select ${a} as res_a

-- can add include anywhere
-- include=easy_sql.sql_processor_test.test_snippet2
-- target=temp.result2
select ${b} as res_b
        """

        steps = StepFactory(None, None).create_from_sql(sql)

        self.assertEqual(4, len(steps))
        self.assertEqual(run_sql(sql, "result1", spark=spark), [(1,)])
        self.assertEqual(run_sql(sql, "result2", spark=spark), [(1,)])

    def test_include_snippet_from_sql(self):
        sql = """
-- can have comments here
-- include=/tmp/test_snippet.sql
-- target=temp.result1
select ${a} as res_a

-- can add include anywhere
-- include=/tmp/test_snippet2.sql
-- target=temp.result2
select ${b} as res_b
        """
        with open("/tmp/test_snippet.sql", "w") as f:
            f.write(test_snippet)
        with open("/tmp/test_snippet2.sql", "w") as f:
            f.write(test_snippet2)

        spark = LocalSpark.get()
        steps = StepFactory(None, None).create_from_sql(sql)

        self.assertEqual(4, len(steps))
        self.assertEqual(run_sql(sql, "result1", spark=spark), [(1,)])
        self.assertEqual(run_sql(sql, "result2", spark=spark), [(1,)])



import unittest

from easy_sql import base_test
from easy_sql.base_test import (
    TEST_BQ_URL,
    TEST_CH_URL,
    TEST_PG_URL,
    bigquery_sql_expr,
    date,
    dt,
    dt_zone,
)


class SqlProcessorTest(unittest.TestCase):
    def test_should_run_sql_for_pg_backend(self):
        from easy_sql.sql_processor.backend.rdb import RdbBackend, _exec_sql

        backend = RdbBackend(TEST_PG_URL)
        _exec_sql(backend.conn, "drop schema if exists t cascade")
        _exec_sql(backend.conn, "create schema t")
        self.run_sql_for_pg_backend(backend)

    def test_should_run_sql_for_ch_backend(self):
        from easy_sql.sql_processor.backend.rdb import RdbBackend, _exec_sql

        backend = RdbBackend(TEST_CH_URL)
        _exec_sql(backend.conn, "drop database if exists t")
        _exec_sql(backend.conn, "create database t")
        self.run_sql_for_pg_backend(backend)

    def test_should_run_sql_for_bq_backend(self):
        if not base_test.should_run_integration_test("bq"):
            return
        import os

        from easy_sql.sql_processor.backend.rdb import RdbBackend, _exec_sql

        backend = RdbBackend(
            TEST_BQ_URL,
            credentials=f"{os.environ.get('HOME', '/tmp')}/.bigquery/credential-test.json",
            sql_expr=bigquery_sql_expr,
        )
        _exec_sql(backend.conn, "drop schema if exists t cascade")
        _exec_sql(backend.conn, "create schema if not exists t")
        self.run_sql_for_pg_backend(backend)

    def test_should_run_sql_for_ch_backend_for_dynamic_partitions(self):
        from easy_sql.sql_processor.backend.rdb import RdbBackend, _exec_sql

        backend = RdbBackend(TEST_CH_URL)
        _exec_sql(backend.conn, "drop database if exists t")
        _exec_sql(backend.conn, "create database t")
        self.run_sql_for_dynamic_partitions(backend)

    def test_should_run_sql_for_pg_backend_for_dynamic_partitions(self):
        from easy_sql.sql_processor.backend.rdb import RdbBackend, _exec_sql

        backend = RdbBackend(TEST_PG_URL)
        _exec_sql(backend.conn, "drop schema if exists t cascade")
        _exec_sql(backend.conn, "create schema t")
        self.run_sql_for_dynamic_partitions(backend)

    def test_should_run_sql_for_bq_backend_for_dynamic_partitions(self):
        if not base_test.should_run_integration_test("bq"):
            return
        import os

        from easy_sql.sql_processor.backend.rdb import RdbBackend, _exec_sql

        backend = RdbBackend(
            TEST_BQ_URL,
            credentials=f"{os.environ.get('HOME', '/tmp')}/.bigquery/credential-test.json",
            sql_expr=bigquery_sql_expr,
        )
        _exec_sql(backend.conn, "drop schema if exists t cascade")
        _exec_sql(backend.conn, "create schema t")
        self.run_sql_for_dynamic_partitions(backend)

    def run_sql_for_dynamic_partitions(self, backend):
        from easy_sql.sql_processor import SqlProcessor

        prefix = "${temp_db}." if backend.is_bigquery_backend else ""
        sql = f"""
        -- target=variables
        select '2021-01-01'       as __partition__data_date
        -- target=temp.result
        select 1 as a
        -- target=output.t.result
        select *, 2 as b from {prefix}result
        """
        processor = SqlProcessor(
            backend,
            sql,
            variables={
                "__create_output_table__": True,
                "temp_db": backend.temp_schema if backend.is_bigquery_backend else None,
            },
        )
        processor.func_runner.register_funcs({"t": lambda a, b: int(a) + int(b)})
        processor.run(dry_run=False)
        sql = f"""
        -- target=temp.result1
        select 1 as a
        -- target=output.t.result
        select *, 3 as b, '2021-01-01' as data_date from {prefix}result1 union all
        select *, 3 as b, '2021-01-02' as data_date from {prefix}result1
        -- target=output.t.result
        select *, 3 as b, '2021-01-03' as data_date from {prefix}result1
        """
        processor = SqlProcessor(
            backend, sql, variables={"temp_db": backend.temp_schema if backend.is_bigquery_backend else None}
        )
        processor.func_runner.register_funcs({"t": lambda a, b: int(a) + int(b)})
        processor.run(dry_run=False)

        if backend.is_bigquery_backend:
            mock_dt_1, mock_dt_2, mock_dt_3 = date("2021-01-01"), date("2021-01-02"), date("2021-01-03")
        else:
            mock_dt_1, mock_dt_2, mock_dt_3 = "2021-01-01", "2021-01-02", "2021-01-03"
        self.assertEqual(
            backend.exec_sql("select * from t.result order by data_date").collect(),
            [(1, 3, mock_dt_1), (1, 3, mock_dt_2), (1, 3, mock_dt_3)],
        )

    def run_sql_for_pg_backend(self, backend):
        from easy_sql.sql_processor import SqlProcessor

        prefix = "${temp_db}." if backend.is_bigquery_backend else ""
        sql = f"""
        -- backend: postgres
        -- target=variables
        select
            1                    as __create_output_table__
            , 'append'           as __save_mode__
            , '2021-01-01'       as __partition__data_date
        -- target=temp.result
        select 1 as a
        -- target=temp.result1
        select *, 2 as b from {prefix}result
        -- target=output.t.result
        select *, 2 as b from {prefix}result
        -- target=output.t.result
        select * from {prefix}result1
        -- target=output.t.result1
        select 1 as a, 2 as b, 't' as c, cast('2021-01-01' as timestamp) as d, ${{t(1, 2)}} as e
        -- target=output.t.result1
        select 1 as a, 2 as b, 't' as c, cast('2021-01-01' as timestamp) as d, ${{t(1, 2)}} as e
        """
        processor = SqlProcessor(
            backend,
            sql,
            variables={
                "__create_output_table__": True,
                "temp_db": backend.temp_schema if backend.is_bigquery_backend else None,
            },
        )
        processor.func_runner.register_funcs({"t": lambda a, b: int(a) + int(b)})
        processor.run(dry_run=False)

        if backend.is_bigquery_backend:
            from datetime import timedelta, timezone

            mock_dt, mock_datetime = date("2021-01-01"), dt_zone(
                "2021-01-01 00:00:00", timezone=timezone(timedelta(hours=0))
            )
        else:
            mock_dt, mock_datetime = "2021-01-01", dt("2021-01-01 00:00:00")

        self.assertEqual(backend.exec_sql("select * from t.result").collect(), [(1, 2, mock_dt), (1, 2, mock_dt)])
        self.assertEqual(
            backend.exec_sql("select * from t.result1").collect(),
            [(1, 2, "t", mock_datetime, 3, mock_dt), (1, 2, "t", mock_datetime, 3, mock_dt)],
        )



from __future__ import annotations

import traceback
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, List, Optional, Union

from ..logger import logger
from . import SqlProcessorException

if TYPE_CHECKING:
    from .backend import Backend
    from .context import ProcessorContext
    from .step import Step


__all__ = ["ColumnFuncs", "TableFuncs", "PartitionFuncs", "AlertFunc"]


class ColumnFuncs:
    def __init__(self, backend: Backend):
        self.backend = backend

    def all_cols_without_one_expr(self, table_name: str, *cols_to_exclude: str) -> str:
        return self.all_cols_with_exclusion_expr(table_name, *cols_to_exclude)

    def all_cols_with_exclusion_expr(self, table_name: str, *cols_to_exclude: str) -> str:
        pure_table_name = table_name.split(".")[1] if "." in table_name else table_name
        fields = self.backend.exec_sql(f"select * from {table_name} limit 0").field_names()
        return ", ".join(
            [
                f"{pure_table_name}.{col}"
                for col in fields
                if col not in cols_to_exclude or (col.find(".") != -1 and col.split(".")[-1] not in cols_to_exclude)
            ]
        )

    def all_cols_prefixed_with_exclusion_expr(self, table_name: str, prefix: str, *cols_to_exclude: str) -> str:
        # this function not support pg yet
        pure_table_name = table_name.split(".")[1] if "." in table_name else table_name
        fields = self.backend.exec_sql(f"select * from {table_name} limit 0").field_names()
        return ", ".join(
            [
                f"{pure_table_name}.{col} as `{prefix}{col}`"
                for col in fields
                if col not in cols_to_exclude or (col.find(".") != -1 and col.split(".")[-1] not in cols_to_exclude)
            ]
        )


class TableFuncs:
    def __init__(self, backend: Backend):
        self.backend = backend

    def ensure_no_null_data_in_table(self, step: Step, table_name: str, query: str = None) -> bool:
        fields = self.backend.exec_sql(f"select * from {table_name} limit 0").field_names()
        return self._check_not_null_columns_in_table(step, table_name, fields, query, "ensure_no_null_data_in_table")

    def check_not_null_column_in_table(
        self, step: Step, table_name: str, not_null_column: str, query: str = None
    ) -> bool:
        return self._check_not_null_columns_in_table(
            step, table_name, [not_null_column], query, "check_not_null_column_in_table"
        )

    def _check_not_null_columns_in_table(
        self,
        step: Step,
        table_name: str,
        not_null_columns: List[str],
        query: str = None,
        context: str = "check_not_null_column_in_table",
    ) -> bool:
        null_counts = {}
        for field in not_null_columns:
            condition_sql = f"{field} is null"
            sql = f"select count(1) from {table_name} where " + (
                condition_sql if query is None else f"({condition_sql}) and ({query})"
            )
            null_count = self.backend.exec_sql(sql).collect()[0][0]
            null_counts[field] = null_count
        if sum(null_counts.values()) != 0:
            null_count_msg = "\n".join([f"{v} null rows({k})" for k, v in null_counts.items() if v != 0])
            msg = f"{context} {table_name} failed, found: \n{null_count_msg}"
            logger.info(msg)
            step.collect_report(message=msg)
            return False
        else:
            return True


class PartitionFuncs:
    def __init__(self, backend: Backend):
        self.backend = backend

    def is_not_first_partition(self, table_name: str, partition_value: str) -> bool:
        return not self.is_first_partition(table_name, partition_value)

    def is_first_partition(self, table_name: str, partition_value: str) -> bool:
        partition_values = self._get_partition_values(table_name)
        return partition_value == str(partition_values[0]) if len(partition_values) > 0 else False

    def _get_partition_values(self, table_name) -> List[Union[str, int]]:
        raise NotImplementedError()

    def _get_partition_values_as_str(self, table_name) -> List[str]:
        return [str(v) for v in self._get_partition_values(table_name)]

    def partition_exists(self, table_name: str, partition_value: str) -> bool:
        return partition_value in self._get_partition_values_as_str(table_name)

    def ensure_partition_exists(self, step, *args) -> bool:
        step: Step = step
        if len(args) < 2:
            raise Exception(
                "must contains at least one table and exactly one partition_value when calling"
                f" ensure_partition_exists, got {args}"
            )
        partition_value = args[-1]
        tables = args[:-1]
        partition_not_exists_tables = []
        for table in tables:
            try:
                if not self.partition_exists(table, partition_value):
                    partition_not_exists_tables.append(table)
            except Exception:
                partition_not_exists_tables.append(table)
        if len(partition_not_exists_tables) > 0:
            message = f"partition {partition_value} not exists: {partition_not_exists_tables}"
            logger.info(message)
            step.collect_report(message=message)
            return False
        return True

    def ensure_dwd_partition_exists(self, step, *args) -> bool:
        step: Step = step
        if len(args) < 2:
            raise Exception(
                "must contains one table and exactly one partition_value and one or more foreign key columns"
                f" when calling ensure_dwd_partition_exists, got {args}"
            )
        table_name = args[0]
        partition_value = args[1]
        foreign_key_cols = args[2:]
        check_ok = True
        try:
            if not self.partition_exists(table_name, partition_value):
                first_partition = self.get_first_partition_optional(table_name)
                if not first_partition or partition_value > first_partition:
                    message = f"partition {partition_value} not exists: {table_name}"
                    logger.info(message)
                    step.collect_report(message=message)
                    check_ok = False
        except Exception:
            message = f"partition {partition_value} not exists: {table_name}"
            logger.info(message)
            step.collect_report(message=message)
            check_ok = False

        if len(foreign_key_cols) > 0:
            partition_values = self._get_partition_values(table_name)
            if len(partition_values) > 0:  # this indicates first partition exists
                partition_col, first_partition = self.get_partition_col(table_name), str(partition_values[0])
                partition_value_to_use = partition_value if partition_value > first_partition else first_partition
                partition_value_to_use_expr = (
                    f"'{partition_value_to_use}'" if isinstance(partition_values[0], str) else partition_value_to_use
                )
                sql = (
                    f"select count(1) as total_count from {table_name} where"
                    f" {partition_col}={partition_value_to_use_expr}"
                )
                total_count = self.backend.exec_sql(sql).collect()[0][0]
                if total_count > 0:
                    fk_col_non_null_expr = " or ".join([f"{fk_col} is not null" for fk_col in foreign_key_cols])
                    sql = (
                        f"select 1 as nonnull_count from {table_name} "
                        f"where ({fk_col_non_null_expr}) and {partition_col}={partition_value_to_use_expr} limit 1"
                    )
                    any_fk_col_non_null_rows = self.backend.exec_sql(sql).collect()
                    if len(any_fk_col_non_null_rows) == 0:
                        message = (
                            f"all fk cols are null in partition: table_name={table_name},"
                            f" partition={partition_value_to_use}"
                        )
                        logger.info(message)
                        step.collect_report(message=message)
                        check_ok = False

        return check_ok

    def ensure_partition_or_first_partition_exists(self, step, *args) -> bool:
        step: Step = step
        if len(args) < 2:
            raise Exception(
                "must contains at least one table and exactly one partition_value when calling"
                f" ensure_partition_exists, got {args}"
            )
        partition_value = args[-1]
        tables = args[:-1]
        partition_not_exists_tables = []
        for table in tables:
            try:
                if not self.partition_exists(table, partition_value):
                    first_partition = self.get_first_partition_optional(table)
                    if not first_partition or partition_value > first_partition:
                        partition_not_exists_tables.append(table)
            except Exception:
                partition_not_exists_tables.append(table)
        if len(partition_not_exists_tables) > 0:
            message = f"partition {partition_value} not exists: {partition_not_exists_tables}"
            logger.info(message)
            step.collect_report(message=message)
            return False
        return True

    def partition_not_exists(self, table_name: str, partition_value: str) -> bool:
        return not self.partition_exists(table_name, partition_value)

    def previous_partition_exists(self, table_name: str, curr_partition_value_as_dt: str) -> bool:
        partition_values = self._get_partition_values_as_str(table_name)
        partition_date_format = "%Y-%m-%d" if "-" in curr_partition_value_as_dt else "%Y%m%d"
        try:
            curr_partition_value_as_dt = datetime.strptime(curr_partition_value_as_dt, partition_date_format)
        except ValueError:
            raise SqlProcessorException(
                f"partition value must be date of format `%Y-%m-%d` or `%Y%m%d`, found {curr_partition_value_as_dt}"
            )
        previous_partition_value = (curr_partition_value_as_dt - timedelta(days=1)).strftime(partition_date_format)
        return previous_partition_value in partition_values

    def get_partition_or_first_partition(self, table_name: str, partition_value: str):
        partition_values = self._get_partition_values_as_str(table_name)
        if len(partition_values) == 0:
            return partition_value
        if partition_value in partition_values:
            return partition_value
        return partition_values[0] if partition_value < partition_values[0] else partition_value

    def get_first_partition_optional(self, table_name: str) -> Optional[str]:
        partition_values = self._get_partition_values(table_name)
        return str(partition_values[0]) if len(partition_values) > 0 else None

    def get_first_partition(self, table_name: str) -> str:
        first_partition = self.get_first_partition_optional(table_name)
        if not first_partition:
            raise Exception(f"no partition found for table {table_name}")
        return first_partition

    def get_last_partition(self, table_name: str) -> str:
        partition_values = self._get_partition_values(table_name)
        last_partition = partition_values[-1] if len(partition_values) > 0 else None
        if not last_partition:
            raise Exception(f"no partition found for table {table_name}")
        return str(last_partition)

    def get_partition_cols(self, table_name: str) -> List[str]:
        raise NotImplementedError()

    def get_partition_col(self, table_name: str):
        partition_cols = self.get_partition_cols(table_name=table_name)
        if not partition_cols:
            raise Exception(f"no partition columns found for table {table_name}")
        return partition_cols[0]

    def has_partition_col(self, table_name: str):
        return len(self.get_partition_cols(table_name)) > 0


class Alerter:
    def send_alert(self, msg: str, mentioned_users: str = ""):
        raise NotImplementedError()


class AlertFunc:
    def __init__(self, backend: Backend, alerter: Alerter):
        self.backend = backend
        self.alerter = alerter

    def alert(
        self, step: Step, context, rule_name: str, pass_condition: str, alert_template: str, mentioned_users: str
    ):
        context: ProcessorContext = context
        # Fetch 10 rows at most
        alert_data = self.backend.exec_sql(step.select_sql).limit(10).collect()
        # alert_data = self.spark.sql(step.select_sql).limit(10).collect()

        msg_list = []
        for data in alert_data:
            check_result = data.as_dict()
            context.add_vars(check_result)
            if not step.func_runner.run_func(pass_condition.format(**check_result), context.vars_context):
                msg_list.append(alert_template.format(**check_result))
        if msg_list:
            self.alerter.send_alert("\n".join(msg_list), mentioned_users)

    def alert_exception_handler(self, rule_name: str, mentioned_users: str):
        def exception_handler(e):
            logger.error(f"{traceback.format_exc()}")
            self.alerter.send_alert(f"执行{rule_name}失败: {e}", mentioned_users)

        return exception_handler



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Tuple, Union

if TYPE_CHECKING:
    from pyspark.sql.types import StructType


__all__ = ["Backend", "Table", "Row", "TableMeta", "Partition", "SaveMode"]


class Col:
    def __init__(self, name: str, type: str):
        self.name, self.type = name, type

    def as_dict(self) -> Dict[str, str]:
        return {"name": self.name, "type": self.type}


class Backend:
    @property
    def is_spark_backend(self):
        return str(self.__class__) == "<class 'easy_sql.sql_processor.backend.spark.SparkBackend'>"

    @property
    def is_postgres_backend(self):
        return str(self.__class__) == "<class 'easy_sql.sql_processor.backend.rdb.RdbBackend'>" and self.is_pg

    @property
    def is_clickhouse_backend(self):
        return str(self.__class__) == "<class 'easy_sql.sql_processor.backend.rdb.RdbBackend'>" and self.is_ch

    @property
    def is_bigquery_backend(self):
        return str(self.__class__) == "<class 'easy_sql.sql_processor.backend.rdb.RdbBackend'>" and self.is_bq

    @property
    def is_rdb_backend(self):
        return self.is_postgres_backend or self.is_clickhouse_backend or self.is_bigquery_backend

    def reset(self):
        raise NotImplementedError()

    def init_udfs(self, *args, **kwargs):
        raise NotImplementedError()

    def register_udfs(self, funcs: Dict[str, Callable]):
        raise NotImplementedError()

    def create_empty_table(self):
        raise NotImplementedError()

    def exec_native_sql(self, sql: str) -> Any:
        raise NotImplementedError()

    def exec_sql(self, sql: str) -> Table:
        raise NotImplementedError()

    def temp_tables(self) -> List[str]:
        raise NotImplementedError()

    def clear_cache(self):
        raise NotImplementedError()

    def clear_temp_tables(self, exclude: List[str] = None):
        raise NotImplementedError()

    def create_temp_table(self, table: Table, name: str):
        raise NotImplementedError()

    def create_cache_table(self, table: Table, name: str):
        raise NotImplementedError()

    def broadcast_table(self, table: Table, name: str):
        raise NotImplementedError()

    def table_exists(self, table: TableMeta):
        pass

    def refresh_table_partitions(self, table: TableMeta):
        raise NotImplementedError()

    def save_table(
        self, source_table: TableMeta, target_table: TableMeta, save_mode: SaveMode, create_target_table: bool
    ):
        raise NotImplementedError()

    def clean(self):
        raise NotImplementedError()

    def create_table_with_data(
        self,
        full_table_name: str,
        values: List[List[Any]],
        schema: Union[StructType, List[Col]],
        partitions: List[Partition],
    ):
        raise NotImplementedError()

    def create_temp_table_with_data(
        self, table_name: str, values: List[List[Any]], schema: Union[StructType, List[Col]]
    ):
        raise NotImplementedError()


class Partition:
    def __init__(self, field: str, value=None):
        self.field = field
        self.value = value

    def __repr__(self):
        return f"Partition(field={self.field}, value={self.value}"

    def __str__(self):
        return f"{self.field}={self.value}"

    def __eq__(self, other):
        if not isinstance(other, Partition):
            return False
        return self.field == other.field and self.value == other.value

    def __hash__(self):
        return str(self).__hash__()


class SaveMode(Enum):
    overwrite = (0,)
    append = 1


class TableMeta:
    def __init__(self, table_name: str, partitions: List[Partition] = None):
        self.table_name = table_name
        self.partitions = partitions or []
        self.dbname, self.pure_table_name = self.__parse_table_name()

    def __repr__(self):
        return (
            f"TableMeta(table_name={self.table_name}"
            f", partitions={self.partitions}"
            f", dbname={self.dbname}"
            f", pure_table_name={self.pure_table_name})"
        )

    def update_partitions(self, partitions: List[Partition]) -> TableMeta:
        self.partitions = partitions
        return self

    def clone_with_name(self, table_name: str) -> TableMeta:
        return TableMeta(table_name, self.partitions)

    def clone_with_partitions(self, partitions: List[Partition]) -> TableMeta:
        return TableMeta(self.table_name, partitions)

    def __parse_table_name(self) -> Tuple[str, str]:
        if self.table_name.find(".") != -1:
            if len(self.table_name.split(".")) != 2:
                raise Exception(f"table_name must be like DB_NAME.TABLE_NAME, found: {self.table_name}")
            dbname = self.table_name[: self.table_name.find(".")]
            pure_table_name = self.table_name[self.table_name.find(".") + 1 :]
        else:
            dbname, pure_table_name = None, self.table_name
        return dbname, pure_table_name

    def has_partitions(self):
        return len(self.partitions) > 0

    def has_dynamic_partition(self):
        return any([pt.value is None for pt in self.partitions])

    def get_full_table_name(self, temp_db: str = None):
        return f"{self.dbname or temp_db}.{self.pure_table_name}"


class Table:
    def is_empty(self) -> bool:
        raise NotImplementedError()

    def field_names(self) -> List[str]:
        raise NotImplementedError()

    def first(self) -> Row:
        raise NotImplementedError()

    def limit(self, count: int) -> Table:
        raise NotImplementedError()

    def with_column(self, name: str, value: any) -> Table:
        raise NotImplementedError()

    def collect(self) -> List[Row]:
        raise NotImplementedError()

    def show(self, count: int):
        raise NotImplementedError()

    def count(self) -> int:
        raise NotImplementedError()


class Row:
    def as_dict(self) -> Dict[str, Any]:
        raise NotImplementedError()

    def __str__(self):
        raise NotImplementedError()

    def __getitem__(self, i):
        raise NotImplementedError()

    def as_tuple(self) -> Tuple:
        raise NotImplementedError()



from __future__ import annotations

import os
import re
import urllib.parse
from datetime import datetime
from os import path
from typing import TYPE_CHECKING, Any, Dict, List, Optional

if TYPE_CHECKING:
    from easy_sql.sql_processor.backend import Backend

import click


def resolve_file(file_path: str, abs_path: bool = False) -> str:
    if file_path.lower().startswith("hdfs://"):
        # do not resolve if it is hdfs path
        return file_path
    base_path = os.path.abspath(os.curdir)
    if not path.exists(file_path):
        if path.exists(path.join(base_path, file_path)):
            file_path = path.join(base_path, file_path)
        elif path.exists(path.basename(file_path)):
            file_path = path.basename(file_path)
        else:
            raise Exception(f"file not found: {file_path}")
    if abs_path:
        file_path = path.abspath(file_path)
    if " " in file_path:
        parts = file_path.split("/")
        # Remove space inside file path, since spark will raise issue with this case.
        # We must ensure there is a soft link to the path with space removed to the end.
        file_path = "/".join([re.sub(r" .*$", "", part) for part in parts])
    return file_path


def resolve_files(files_path: str, abs_path: bool = False) -> str:
    return ",".join([resolve_file(f.strip(), abs_path) for f in files_path.split(",") if f.strip()])


def read_sql(sql_file: str):
    with open(resolve_file(sql_file)) as f:
        return f.read()


@click.command(name="data_process")
@click.option("--sql-file", "-f", type=str)
@click.option("--vars", "-v", type=str, required=False)
@click.option("--dry-run", type=str, required=False, help="if dry run, one of [true, 1, false, 0]")
@click.option("--print-command", "-p", is_flag=True)
def data_process(sql_file: str, vars: str, dry_run: str, print_command: bool):
    _data_process(sql_file, vars, dry_run, print_command)


def _data_process(sql_file: str, vars: Optional[str], dry_run: Optional[str], print_command: bool):
    if not sql_file.endswith(".sql"):
        raise Exception(f"sql_file must ends with .sql, found `{sql_file}`")
    dry_run = dry_run if dry_run is not None else "0"

    if print_command:
        command = shell_command(sql_file=sql_file, vars=vars, dry_run=dry_run)
        print(command)
        return command

    dry_run = dry_run in ["true", "1"]
    config = EasySqlConfig.from_sql(sql_file)

    from easy_sql.sql_processor import SqlProcessor

    def run_with_vars(backend: Backend, variables: Dict[str, Any]):
        vars_dict = dict(
            [
                (v.strip()[: v.strip().index("=")], urllib.parse.unquote_plus(v.strip()[v.strip().index("=") + 1 :]))
                for v in vars.split(",")
                if v.strip()
            ]
            if vars
            else []
        )
        variables.update(vars_dict)

        sql_processor = SqlProcessor(
            backend, config.sql, variables=variables, scala_udf_initializer=config.scala_udf_initializer
        )
        if config.udf_file_path:
            sql_processor.register_udfs_from_pyfile(
                resolve_file(config.udf_file_path) if "/" in config.udf_file_path else config.udf_file_path
            )
        if config.func_file_path:
            sql_processor.register_funcs_from_pyfile(
                resolve_file(config.func_file_path) if "/" in config.func_file_path else config.func_file_path
            )

        sql_processor.run(dry_run=dry_run)

    backend: Backend = create_sql_processor_backend(config.backend, config.sql, config.task_name)

    backend_is_bigquery = config.backend == "bigquery"
    pre_defined_vars = {"temp_db": backend.temp_schema if backend_is_bigquery else None}
    try:
        run_with_vars(backend, pre_defined_vars)
    finally:
        backend.clean()


def create_sql_processor_backend(backend: str, sql: str, task_name: str) -> Backend:
    if backend == "spark":
        from easy_sql.spark_optimizer import get_spark
        from easy_sql.sql_processor.backend import SparkBackend

        backend = SparkBackend(get_spark(task_name))
        exec_sql = lambda sql: backend.exec_native_sql(sql)  # noqa: E731
    elif backend == "maxcompute":
        odps_parms = {"access_id": "xx", "secret_access_key": "xx", "project": "xx", "endpoint": "xx"}
        from easy_sql.sql_processor.backend.maxcompute import (
            MaxComputeBackend,
            _exec_sql,
        )

        backend = MaxComputeBackend(**odps_parms)
        exec_sql = lambda sql: _exec_sql(backend.conn, sql)  # noqa: E731
    elif backend in ["postgres", "clickhouse", "bigquery"]:
        from easy_sql.sql_processor.backend.rdb import RdbBackend, _exec_sql

        if backend == "postgres":
            assert "PG_URL" in os.environ, "Must set PG_URL env var to run an ETL with postgres backend."
            backend = RdbBackend(os.environ["PG_URL"])
        elif backend == "clickhouse":
            assert (
                "CLICKHOUSE_URL" in os.environ
            ), "Must set CLICKHOUSE_URL env var to run an ETL with Clickhouse backend."
            backend = RdbBackend(os.environ["CLICKHOUSE_URL"])
        elif backend == "bigquery":
            assert (
                "BIGQUERY_CREDENTIAL_FILE" in os.environ
            ), "Must set BIGQUERY_CREDENTIAL_FILE env var to run an ETL with BigQuery backend."
            backend = RdbBackend("bigquery://", credentials=os.environ["BIGQUERY_CREDENTIAL_FILE"])
        else:
            raise Exception(f"unsupported backend: {backend}")
        exec_sql = lambda sql: _exec_sql(backend.conn, sql)  # noqa: E731
    else:
        raise Exception("Unsupported backend found: " + backend)

    prepare_sql_list = []
    for line in sql.split("\n"):
        if re.match(r"^-- \s*prepare-sql:.*$", line):
            prepare_sql_list.append(line[line.index("prepare-sql:") + len("prepare-sql:") :].strip())
    for prepare_sql in prepare_sql_list:
        exec_sql(prepare_sql)

    return backend


class EasySqlConfig:
    def __init__(
        self,
        sql_file: str,
        sql: str,
        backend: str,
        customized_backend_conf: List[str],
        customized_easy_sql_conf: List[str],
        udf_file_path: str,
        func_file_path: str,
        scala_udf_initializer: str,
    ):
        self.sql_file = sql_file
        self.sql = sql
        self.backend = backend
        self.customized_backend_conf, self.customized_easy_sql_conf = customized_backend_conf, customized_easy_sql_conf
        self.udf_file_path, self.func_file_path = udf_file_path, func_file_path
        self.scala_udf_initializer = scala_udf_initializer

    @staticmethod
    def from_sql(sql_file: str = None, sql: str = None) -> EasySqlConfig:
        assert sql_file is not None or sql is not None, "sql_file or sql must be set"
        sql = read_sql(sql_file) if sql_file else sql
        sql_lines = sql.split("\n")

        backend = _parse_backend(sql)

        customized_backend_conf: List[str] = []
        customized_easy_sql_conf: List[str] = []
        for line in sql_lines:
            if re.match(r"^-- \s*config:.*$", line):
                config_value = line[line.index("config:") + len("config:") :].strip()
                if config_value.strip().lower().startswith("easy_sql."):
                    customized_easy_sql_conf += [config_value[len("easy_sql.") :]]
                else:
                    customized_backend_conf += [config_value]

        udf_file_path, func_file_path, scala_udf_initializer = None, None, None
        for c in customized_easy_sql_conf:
            if c.startswith("udf_file_path"):
                udf_file_path = c[c.index("=") + 1 :].strip()
            if c.startswith("func_file_path"):
                func_file_path = c[c.index("=") + 1 :].strip()
            if c.startswith("scala_udf_initializer"):
                scala_udf_initializer = c[c.index("=") + 1 :].strip()
        return EasySqlConfig(
            sql_file,
            sql,
            backend,
            customized_backend_conf,
            customized_easy_sql_conf,
            udf_file_path,
            func_file_path,
            scala_udf_initializer,
        )

    @property
    def spark_submit(self):
        # 默认情况下使用系统中默认spark版本下的spark-submit
        # sql代码中指定了easy_sql.spark_submit时，优先级高于default配置
        spark_submit = "spark-submit"
        for c in self.customized_easy_sql_conf:
            if c.startswith("spark_submit"):
                spark_submit = c[c.index("=") + 1 :].strip()
        return spark_submit

    @property
    def task_name(self):
        sql_name = path.basename(self.sql_file)[:-4]
        return f'{sql_name}_{datetime.now().strftime("%Y%m%d%H%M%S")}'

    def spark_conf_command_args(self) -> List[str]:
        # config 的优先级：1. sql 代码里的 config 优先级高于这里的 default 配置
        # 对于数组类的 config，sql 代码里的 config 会添加进来，而不是覆盖默认配置
        default_conf = [
            "spark.master=local[2]",
            "spark.submit.deployMode=client",
            f"spark.app.name={self.task_name}",
            "spark.sql.warehouse.dir=/tmp/spark-warehouse-localdw",
            'spark.driver.extraJavaOptions="-Dderby.system.home=/tmp/spark-warehouse-metastore'
            ' -Dderby.stream.error.file=/tmp/spark-warehouse-metastore.log"',
            f'spark.files="{resolve_file(self.sql_file, abs_path=True)}'
            f'{"," + resolve_file(self.udf_file_path, abs_path=True) if self.udf_file_path else ""}'
            f'{"," + resolve_file(self.func_file_path, abs_path=True) if self.func_file_path else ""}'
            '"',
        ]
        customized_conf_keys = [c[: c.index("=")] for c in self.customized_backend_conf]
        customized_backend_conf = self.customized_backend_conf.copy()

        args = []
        for conf in default_conf:
            conf_key = conf[: conf.index("=")].strip()
            if conf_key not in customized_conf_keys:
                args.append(f"--conf {conf}")
            else:
                customized_conf = [conf for conf in customized_backend_conf if conf.startswith(conf_key)][0]
                if conf_key in ["spark.files", "spark.jars", "spark.submit.pyFiles"]:
                    customized_values = [
                        resolve_file(val.strip(), abs_path=True)
                        for val in customized_conf[customized_conf.index("=") + 1 :].strip('"').split(",")
                        if val.strip()
                    ]
                    default_values = [v for v in conf[conf.index("=") + 1 :].strip('"').split(",") if v]
                    args.append(f'--conf {conf_key}="{",".join(set(default_values + customized_values))}"')
                else:
                    args.append(f"--conf {customized_conf}")
                customized_backend_conf.remove(customized_conf)

        args += [f"--conf {c}" for c in customized_backend_conf]
        return args


def shell_command(sql_file: str, vars: str, dry_run: str):
    config = EasySqlConfig.from_sql(sql_file)

    if config.backend == "spark":
        return (
            f'{config.spark_submit} {" ".join(config.spark_conf_command_args())} "{path.abspath(__file__)}" '
            f"-f {sql_file} --dry-run {dry_run} "
            f'{"-v " + vars if vars else ""} '
        )
    else:
        raise Exception("No need to print shell command for non-spark backends")


def _parse_backend(sql: str):
    sql_lines = sql.split("\n")

    backend = "spark"
    for line in sql_lines:
        if re.match(r"^-- \s*backend:.*$", line):
            backend = line[line.index("backend:") + len("backend:") :].strip().split(" ")[0]
            break
    supported_backends = ["spark", "postgres", "clickhouse", "maxcompute", "bigquery"]
    if backend not in supported_backends:
        raise Exception(f"unsupported backend `${backend}`, all supported backends are: {supported_backends}")
    return backend


if __name__ == "__main__":
    # test cases:
    # print(shell_command('source/dm/dm_source.sales_count.spark.sql', '20210505', 'day', '1', 'test-task', 'test-task', None, None, None)) # noqa: B950
    # print(shell_command('sales/dm/indicator/sql/dm_sales.order_count.sql', '20210505', 'day', '1', 'test-task', 'test-task',
    #                     None, None, None, None, None, '1').replace('--conf', '\n --conf'))
    # print(shell_command('sales/samples/dm_source.sales_count.ch.sql', '20210505', 'day', '1', 'test-task', 'test-task',
    #                     None, None, None, None, '1').replace('--conf', '\n --conf'))
    # print(shell_command('sales/samples/aws_glue_sample.sql', '20210505', 'day', '1', 'test-task', 'test-task', None,
    #                     'sales/etl_generator.py', 'sales/samples/aws_funcs.py', '', '1').replace('--conf', '\n --conf'))
    # print(shell_command('sales/samples/dm_source.sales_count.mc.sql', '20220301', 'day', '1', 'test-task', 'test-task',
    #                     None, None, None, None, '1').replace('--conf', '\n --conf'))
    data_process()



import unittest

from easy_sql.data_process import EasySqlConfig


class EasySqlConfigTest(unittest.TestCase):
    def test_parse_config(self):
        config = EasySqlConfig.from_sql(
            sql="""
-- config: easy_sql.udf_file_path=test/sample_data_process.py
-- config: easy_sql.func_file_path=test/sample_data_process.py
-- config: easy_sql.spark_submit=/my/custom/spark-submit
-- config: spark.files=test/sample_etl.spark.sql,test/sample_etl.postgres.sql,
        """,
            sql_file="",
        )
        self.assertEqual(config.spark_submit, "/my/custom/spark-submit")
        print(config.spark_conf_command_args())
        self.assertEqual(config.spark_conf_command_args()[-1].count(","), 3)
        self.assertEqual(config.spark_submit, "/my/custom/spark-submit")



from typing import Dict

from pyspark.sql import SparkSession


class SparkDynamicConfig:
    def __init__(self, max_shuffle_partitions: int = None, min_shuffle_partitions: int = None):
        self.max_shuffle_partitions = max_shuffle_partitions
        self.min_shuffle_partitions = min_shuffle_partitions

    def use_min_shuffle_partitions(self, spark: SparkSession) -> "SparkDynamicConfig":
        assert self.min_shuffle_partitions, "must provide min_shuffle_partitions to use the conf"
        spark.conf.set("spark.sql.adaptive.enabled", "false")
        spark.conf.set("spark.sql.shuffle.partitions", str(self.min_shuffle_partitions))
        spark.conf.set("spark.default.parallelism", str(self.min_shuffle_partitions))
        return self

    def use_max_shuffle_partitions(self, spark: SparkSession) -> "SparkDynamicConfig":
        assert self.max_shuffle_partitions, "must provide max_shuffle_partitions to use the conf"
        spark.conf.set("spark.sql.adaptive.enabled", "false")
        spark.conf.set("spark.sql.shuffle.partitions", str(self.max_shuffle_partitions))
        spark.conf.set("spark.default.parallelism", str(self.max_shuffle_partitions))
        return self

    def use_adaptive_shuffle_partitions(self, spark: SparkSession) -> "SparkDynamicConfig":
        spark.conf.set("spark.sql.adaptive.enabled", "true")
        return self


def get_spark(app_name: str, conf: Dict = None):
    builder = SparkSession.builder.appName(app_name).enableHiveSupport()
    conf = conf or {}
    for k, v in conf.items():
        builder.config(k, v)

    spark = builder.getOrCreate()
    spark.conf.set("spark.sql.statistics.fallBackToHdfs", "true")
    # 启用 Adaptive Execution ，从而启用自动设置 Shuffle Reducer 特性
    spark.conf.set("spark.sql.adaptive.enabled", "true")
    # 设置每个 Reducer 读取的目标数据量，单位为字节。默认64M，一般改成集群块大小
    spark.conf.set("spark.sql.adaptive.shuffle.targetPostShuffleInputSize", "134217728")
    # 允许动态资源分配，配合 spark.dynamicAllocation.minExecutors, spark.dynamicAllocation.maxExecutors 等使用
    # spark 3.0+ 不允许动态设置以下两个参数
    import pyspark

    if str(pyspark.__version__).startswith("2."):
        spark.conf.set("spark.dynamicAllocation.enabled", "true")
        spark.conf.set("spark.shuffle.service.enabled", "true")

    # spark.conf.set("hive.exec.dynamic.partition", "true")
    # default strict. In strict mode, the user must specify at least one static partition,
    # in case the user accidentally overwrites all partitions.
    # In nonstrict mode all partitions are allowed to be dynamic.
    spark.conf.set("hive.exec.dynamic.partition.mode", "nonstrict")

    return spark


def clear_temp_views(spark: SparkSession):
    for table in spark.catalog.listTables("default"):
        if table.isTemporary:
            print(f"dropping temp view {table.name}")
            spark.catalog.dropTempView(table.name)



import os.path
import unittest
from datetime import datetime

from easy_sql.sql_tester import SqlReader, TableColumnTypes, TestDataFile, work_path


class TableColumnTypesTest(unittest.TestCase):
    def test_col_type(self):
        tt = TableColumnTypes(
            {
                "a": {
                    "id": "int",
                    "val": "string",
                    "val_decimal": "decimal(10, 2)",
                    "val_bool": "boolean",
                    "val_map": "map<string, string>",
                    "val_date": "date",
                    "val_unknown": "unknown",
                    "val_complicated": "struct<latest_value:string,first_show_time:timestamp>",
                    "val_list": "array<string>",
                    "val_list_int": "array<int>",
                }
            },
            {"pt": "string", "data_date": "string"},
            "spark",
        )
        self.assertEqual(tt.get_col_type("a", "id"), "int")
        self.assertRaises(Exception, lambda: tt.get_col_type("_not_exist_table", "_"))
        self.assertEqual(tt.get_col_type("a", "pt"), "string")
        self.assertRaises(Exception, lambda: tt.get_col_type("a", "_not_exist_col"))

        self.assertEqual(tt.cast_as_type("a", "id", "1"), ("int", 1))
        self.assertEqual(tt.cast_as_type("a", "id", None), ("int", None))
        self.assertEqual(tt.cast_as_type("a", "val", "1"), ("string", "1"))
        self.assertEqual(tt.cast_as_type("a", "val", "null"), ("string", None))
        self.assertEqual(tt.cast_as_type("a", "pt", "null"), ("string", None))
        self.assertEqual(tt.cast_as_type("a", "data_date", "2021-01-01"), ("string", "2021-01-01"))
        self.assertEqual(tt.cast_as_type("a", "pt", "2021-01-01"), ("string", "2021-01-01"))
        self.assertEqual(tt.cast_as_type("a", "pt", "20210101.0"), ("string", "20210101"))
        self.assertEqual(tt.cast_as_type("a", "val_decimal", "20210101.0"), ("decimal(10, 2)", 20210101.0))
        self.assertEqual(tt.cast_as_type("a", "val_bool", "true"), ("boolean", True))
        self.assertEqual(tt.cast_as_type("a", "val_bool", "False"), ("boolean", False))
        self.assertEqual(tt.cast_as_type("a", "val_bool", "1"), ("boolean", True))
        self.assertEqual(tt.cast_as_type("a", "val_bool", "0"), ("boolean", True))
        self.assertEqual(tt.cast_as_type("a", "val_bool", ""), ("boolean", False))
        self.assertEqual(tt.cast_as_type("a", "val_list", "a|b|c"), ("array<string>", ["a", "b", "c"]))
        self.assertEqual(tt.cast_as_type("a", "val_list_int", "1|2"), ("array<int>", [1, 2]))
        self.assertEqual(tt.cast_as_type("a", "val_date", "2021-01-01"), ("date", "2021-01-01"))
        self.assertEqual(
            tt.cast_as_type("a", "val_complicated", "a|2021-01-01"),
            ("struct<latest_value:string,first_show_time:timestamp>", ("a", "2021-01-01")),
        )
        self.assertRaises(Exception, lambda: tt.cast_as_type("a", "val_map", "some map"))
        self.assertRaises(Exception, lambda: tt.cast_as_type("a", "val_unknown", "something"))

        tt = TableColumnTypes(
            {"a": {"id": "Int32", "val": "String", "val_list": "Array<String>", "unsupported": "Nested(...)"}},
            {"pt": "String"},
            "clickhouse",
        )
        self.assertEqual(tt.get_col_type("a", "pt"), "String")
        self.assertRaises(Exception, lambda: tt.cast_as_type("a", "unsupported", "null"))


class TestCaseParserTest(unittest.TestCase):
    def test_parse_case(self):
        class _SqlReader(SqlReader):
            def find_file_path(self, file_name: str) -> str:
                return file_name

        work_path.set_work_path(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        tdf = TestDataFile("test/sample_etl.syntax.xlsx", _SqlReader())
        cases = tdf.parse_test_cases(TableColumnTypes({}, {}, "spark"))
        self.assertEqual(len(cases), 1)
        case = cases[0]
        self.assertEqual(set(case.includes.keys()), {"some_etl_snippets", "some_etl_snippets1"})
        self.assertEqual(
            case.vars, {"data_date": "2021-01-01", "some_var": 1.0, "some_string_var": 2.0, "some_int_var": 3.0}
        )
        self.assertEqual(case.udf_file_paths, ["test/sample_data_process.py", "not_exist.py"])
        self.assertEqual(case.func_file_paths, ["test/sample_data_process.py"])
        self.assertEqual(len(case.inputs), 1)
        input = case.inputs[0]
        self.assertEqual(input.columns, ["id", "val", "val_date", "data_date"])
        self.assertEqual(input.column_types, ["int", "string", "date", "date"])
        self.assertEqual(input.values, [[1, "1.0", datetime(2021, 1, 1, 0, 0), datetime(2021, 1, 1, 0, 0)]])
        self.assertEqual(
            case.outputs[0].values, [[1, "1.0", datetime(2021, 1, 1, 0, 0)], [1, "2.0", None], [1, "3.0", None]]
        )



import codecs
import logging
import sys
from datetime import datetime
from functools import wraps
from typing import Callable

LOG_LEVEL = logging.DEBUG


def _config_logger():
    logger = logging.getLogger("simple_logger")
    logger.setLevel(LOG_LEVEL)
    python_version = sys.version_info
    if python_version.major == 3 and python_version.minor == 6:
        sys.stdout = codecs.getwriter("utf-8")(sys.stdout.detach())
    elif hasattr(sys.stdout, "reconfigure"):
        sys.stdout.reconfigure(encoding="utf-8")
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(LOG_LEVEL)

    formatter = logging.Formatter(
        "[%(asctime)s][%(processName)s:%(threadName)s][%(levelname)s][%(module)s.%(funcName)s:%(lineno)d] %(message)s"
    )
    handler.setFormatter(formatter)

    for existing_handler in logger.handlers:
        logger.removeHandler(existing_handler)
    logger.addHandler(handler)

    return logger


def log_time(func: Callable):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = datetime.now()
        try:
            return func(*args, **kwargs)
        finally:
            end_time = datetime.now()
            logger.debug("function {} took {}s".format(func.__name__, (end_time - start_time).total_seconds()))

    return wrapper


logger = _config_logger()



from __future__ import annotations

import codecs
import logging
import sys
from typing import TYPE_CHECKING, List, Union

import colorlog

if TYPE_CHECKING:
    from sqlfluff.core import SQLBaseError


class LintReporter:
    def __init__(self):
        self.sql_linter_log = LintReporter._create_logger(logging.DEBUG)

    def _get_extra_default_dict(self):
        return {"pos_info": "", "description": "", "warn": "", "pass": ""}

    @staticmethod
    def _create_logger(log_level: Union[int, str]):
        logger = logging.getLogger("linter_logger")
        logger.setLevel(log_level)
        info_formater = colorlog.ColoredFormatter(
            fmt="%(white)s%(message)s%(red)s%(warn)s %(green)s%(pass)s %(blue)s%(pos_info)s %(white)s%(description)s "
        )
        python_version = sys.version_info
        if python_version.major == 3 and python_version.minor == 6:
            sys.stdout = codecs.getwriter("utf-8")(sys.stdout.detach())
        elif hasattr(sys.stdout, "reconfigure"):
            sys.stdout.reconfigure(encoding="utf-8")
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(info_formater)
        for existing_handler in logger.handlers:
            logger.removeHandler(existing_handler)
        logger.addHandler(handler)
        return logger

    def report_violation(self, violation: SQLBaseError, step_start_line=0):
        pos_info = "L: {} | P: {}: | {}  :".format(
            violation.line_no + step_start_line, violation.line_pos, violation.rule_code()
        )
        extra_dict = self._get_extra_default_dict()
        extra_dict["pos_info"] = pos_info
        extra_dict["description"] = violation.desc()
        self.sql_linter_log.info("", extra=extra_dict)

    def report_list_of_violations(self, lint_result: List[SQLBaseError], step_start_line=0):
        if len(lint_result) > 0:
            self.report_warning("Fail")
            for violation in lint_result:
                self.report_violation(violation, step_start_line)
        else:
            self.report_pass("Pass")

    def report_message(self, message):
        self.sql_linter_log.info(message, extra=self._get_extra_default_dict())

    def report_warning(self, warning: str):
        extra_dict = self._get_extra_default_dict()
        extra_dict["warn"] = warning
        self.sql_linter_log.warning("", extra=extra_dict)

    def report_pass(self, pass_info: str):
        extra_dict = self._get_extra_default_dict()
        extra_dict["pass"] = pass_info
        self.sql_linter_log.warning("", extra=extra_dict)



import unittest

from easy_sql.sql_processor import SqlProcessorException, StepConfig


class StepConfigTest(unittest.TestCase):
    def test_should_parse_config(self):
        self.assertEqual(StepConfig.from_config_line("-- target=check.f1", 0), StepConfig("check", "f1", None, 0))
        self.assertEqual(
            StepConfig.from_config_line("-- target=check.f1(a, ${b})", 0), StepConfig("check", "f1(a, ${b})", None, 0)
        )
        self.assertEqual(
            StepConfig.from_config_line("-- target=check.f1(a, ${b}), if=f2(c, ${d})", 0),
            StepConfig("check", "f1(a, ${b})", "f2(c, ${d})", 0),
        )
        self.assertEqual(
            StepConfig.from_config_line("-- target=check.f1(a, ${b}),if=f2(c, ${d})", 0),
            StepConfig("check", "f1(a, ${b})", "f2(c, ${d})", 0),
        )
        self.assertEqual(StepConfig.from_config_line("-- target=variables", 0), StepConfig("variables", None, None, 0))
        self.assertEqual(
            StepConfig.from_config_line("-- target=variables, if=f2(c, ${d})", 0),
            StepConfig("variables", None, "f2(c, ${d})", 0),
        )
        with self.assertRaises(expected_exception=SqlProcessorException):
            StepConfig.from_config_line("-- target=check.f1(a, ${b}),if=f2-(c, ${d})", 0)
        with self.assertRaises(expected_exception=SqlProcessorException):
            StepConfig.from_config_line("-- target=unknown_type", 0)


if __name__ == "__main__":
    unittest.main()



import unittest

from easy_sql.sql_processor.backend import Partition
from easy_sql.sql_processor.backend.sql_dialect import SqlExpr
from easy_sql.sql_processor.backend.sql_dialect.clickhouse import ChSqlDialect


class RdbTest(unittest.TestCase):
    def test_ch_config(self):
        ch_config = ChSqlDialect(SqlExpr(), "dataplat.__table_partitions__")
        sql = ch_config.delete_partition_sql("test.test", [Partition("dt", "20210101")])
        self.assertEqual(
            sql,
            [
                "alter table test.test drop partition tuple('20210101')",
                "alter table dataplat.__table_partitions__ delete "
                "where db_name = 'test' and table_name = 'test' and partition_value = '20210101'",
            ],
        )



import unittest
from unittest.mock import patch

from sqlalchemy.dialects.postgresql import DOUBLE_PRECISION
from sqlalchemy.engine import Inspector, create_mock_engine

from easy_sql.sql_processor.backend.rdb import RdbBackend


class RdbTest(unittest.TestCase):
    def test_get_column_names_should_only_get_from_name(self):
        cols = [{"name": "a"}, {"name": "b"}, {"type": "c"}]
        with patch.object(RdbBackend, "get_columns", return_value=cols):
            rdb = RdbBackend("")
            names = rdb.get_column_names("test")
            self.assertSequenceEqual(names, ["a", "b"])

    def test_get_columns_should_compile_type_by_dialect_when_now_raw(self):
        mock_engine = create_mock_engine("postgresql://", None)
        mock_engine.close = lambda: None
        col = {"name": "id", "type": DOUBLE_PRECISION(10)}
        raw_cols = [col]
        with patch.object(Inspector, "get_columns", return_value=[col.copy() for col in raw_cols]):
            rdb = RdbBackend("")
            rdb.engine = mock_engine

            cols = rdb.get_columns("test")

            self.assertNotEqual(str(col["type"]), "DOUBLE PRECISION")
            self.assertEqual(cols, [{"name": "id", "type": "DOUBLE PRECISION"}])

    def test_get_columns_should_compile_type_by_dialect_when_in_raw(self):
        mock_engine = create_mock_engine("postgresql://", None)
        mock_engine.close = lambda: None
        col = {"name": "id", "type": DOUBLE_PRECISION(10)}
        raw_cols = [col]
        with patch.object(Inspector, "get_columns", return_value=[col.copy() for col in raw_cols]):
            rdb = RdbBackend("")
            rdb.engine = mock_engine

            cols = rdb.get_columns("test", raw=True)

            self.assertEqual(cols, raw_cols)



from __future__ import annotations

import copy
import re
from typing import TYPE_CHECKING, Any, Dict, List

from ..logger import logger

if TYPE_CHECKING:
    from . import FuncRunner

from .common import Column, SqlProcessorException, VarsReplacer

__all__ = ["VarsContext", "TemplatesContext", "ProcessorContext"]


class CommentSubstitutor:
    COMMENT_IDENTIFIABLE_NAME = "__COMMENT_SUBSTITUTED__"

    def __init__(self):
        self.recognized_comment = []

    def substitute(self, sql_expr: str) -> str:
        if CommentSubstitutor.COMMENT_IDENTIFIABLE_NAME in sql_expr:
            raise Exception(
                "Cannot handle sql expression with "
                f"comment identifiable name({CommentSubstitutor.COMMENT_IDENTIFIABLE_NAME}) inside: {sql_expr}"
            )
        lines = []
        recognized_comment = []
        for line in sql_expr.split("\n"):
            if line.startswith("-- "):
                lines.append(CommentSubstitutor.COMMENT_IDENTIFIABLE_NAME + str(len(recognized_comment)) + "__")
                recognized_comment.append(line)
                continue

            current_index = 0
            while True:
                m = re.match(r".*?([^a-zA-Z0-9_]-- ).*", line[current_index:])
                if m:
                    comment_start = m.start(1) + 1
                    left_of_comment = line[: current_index + comment_start]
                    if CommentSubstitutor.is_quote_closed(left_of_comment):
                        lines.append(
                            left_of_comment
                            + CommentSubstitutor.COMMENT_IDENTIFIABLE_NAME
                            + str(len(recognized_comment))
                            + "__"
                        )
                        recognized_comment.append(line[current_index + comment_start :])
                        break
                    else:
                        current_index += comment_start
                else:
                    lines.append(line)
                    break

        self.recognized_comment = recognized_comment
        return "\n".join(lines)

    @staticmethod
    def is_quote_closed(sql_expr: str) -> bool:
        sql_expr = sql_expr.replace("\\\\", "")

        def find_char_without_escape(text: str, char: str) -> int:
            start = 0
            while True:
                idx = text.find(char, start)
                if idx != -1:
                    if idx > 0 and text[idx - 1] == "\\":
                        start = idx + 1
                    else:
                        return idx
                else:
                    return -1

        def _is_quote_closed(quote_index: int, quote_char: str) -> bool:
            start_index = quote_index + 1
            char_without_escape_idx = find_char_without_escape(sql_expr[start_index:], quote_char)
            if char_without_escape_idx == -1:
                return False
            else:
                return CommentSubstitutor.is_quote_closed(sql_expr[start_index + char_without_escape_idx + 1 :])

        # a valid sql_expr should not starts with \' or \", so we don't need to check \\ when find quote start point
        single_quote_index = sql_expr.find("'")
        double_quote_index = sql_expr.find('"')
        if single_quote_index != -1 and double_quote_index == -1:
            return _is_quote_closed(single_quote_index, "'")
        elif single_quote_index != -1 and double_quote_index != -1:
            if single_quote_index < double_quote_index:
                return _is_quote_closed(single_quote_index, "'")
            else:
                return _is_quote_closed(double_quote_index, '"')
        elif single_quote_index == -1 and double_quote_index == -1:
            return True
        elif single_quote_index == -1 and double_quote_index != -1:
            return _is_quote_closed(double_quote_index, '"')

    def recover(self, substituted_sql_expr: str) -> str:
        lines = []
        comment_index = 0
        for line in substituted_sql_expr.split("\n"):
            identifiable_name = f"{CommentSubstitutor.COMMENT_IDENTIFIABLE_NAME}{comment_index}__"
            identifiable_name_count = line.count(identifiable_name)
            if identifiable_name_count > 1:
                raise Exception(f"found multiple comment identifiable name {identifiable_name} in line: {line}")
            if identifiable_name_count == 1 and not line.endswith(identifiable_name):
                raise Exception(
                    f"found comment identifiable name {identifiable_name}, but is not at end in line: {line}"
                )
            if identifiable_name_count == 1:
                lines.append(line.replace(identifiable_name, self.recognized_comment[comment_index]))
                comment_index += 1
            else:
                lines.append(line)
        return "\n".join(lines)


class VarsContext(VarsReplacer):
    def __init__(self, vars: Dict[str, Any] = None, list_vars: Dict[str, List] = None, debug_log: bool = False):
        vars = vars or {}
        self.vars = copy.deepcopy(vars)
        self.list_vars = list_vars or {}
        self.func_runner = None
        self.debug_log = debug_log

    def init(self, func_runner: FuncRunner):
        self.func_runner = func_runner

    def replace_variables(self, text: str, include_funcs: bool = True) -> str:
        m = re.match(r"^\${([^}]+)}$", text.strip())
        if m:
            return self.vars.get(m.group(1).strip(), self.vars.get(m.group(1).strip().lower()))

        original_text = text
        comment_substitutor = CommentSubstitutor()
        text = comment_substitutor.substitute(text)

        variables = self.vars
        for key in variables.keys():
            text = re.sub(re.escape(f"${{{key}}}"), str(variables.get(key)), text, flags=re.IGNORECASE)
        self._log_replace_process(f"after direct variable replaced: {text}")

        if not include_funcs:
            return comment_substitutor.recover(text)

        var_rex = re.compile(r"\${([^}]+)}")
        text_parts = []
        match = None
        while True:
            start = match.end() if match is not None else 0
            match = var_rex.search(text, start)
            if match is None:
                text_parts.append(text[start:])
                break

            text_parts.append(text[start : match.start()])
            var_name = var_rex.match(match.group()).groups()[0]
            var_name_is_func = "(" in var_name
            self._log_replace_process(f"variable matched: var_name={var_name}, is_func={var_name_is_func}")
            if var_name_is_func:
                var_value = self.func_runner.run_func(var_name, self)
            elif var_name in variables:
                var_value = variables[var_name]
            elif var_name.upper() in variables:
                var_value = variables[var_name.upper()]
            else:
                raise SqlProcessorException(
                    f"unknown variable `{var_name}`. text={original_text}, known_vars={variables}"
                )

            text_parts.append(str(var_value))
            self._log_replace_process(f"var_value: {var_value}")

        text = "".join(text_parts)
        text = comment_substitutor.recover(text)
        self._log_replace_process(f"after variable replaced: text={text}")
        return text

    def _log_replace_process(self, message: str):
        if self.debug_log:
            logger.debug(message)

    def add_vars(self, vars: Dict[str, str]):
        vars = {k.lower(): v for k, v in vars.items()}
        self.vars.update(vars)
        if self.debug_log:
            logger.debug(f"vars: {self.vars}")

    def add_list_vars(self, vars: Dict[str, List]):
        vars = {k.lower(): v for k, v in vars.items()}
        self.list_vars.update(vars)
        if self.debug_log:
            logger.debug(f"list vars: {self.vars}")


class TemplatesContext:
    def __init__(self, debug_log: bool = False, templates: dict = None):
        self.templates: Dict[str, str] = templates or {}
        self.debug_log = debug_log

    def replace_templates(self, text: str):
        templates = self.templates
        tmpl_with_arg_pattern = re.compile(r"@{\s*(\w+)\(\s*?(\s*\w+\s*=\s*[^,)]+\s*,?\s*)*\)\s*}", flags=re.IGNORECASE)
        tmpl_no_arg_pattern = re.compile(r"@{\s*(\w+)\s*}", flags=re.IGNORECASE)
        # pattern = re.compile(r'(%s)|(%s)' % (tmpl_with_arg_rex, tmpl_no_arg_rex), flags=re.IGNORECASE)
        while tmpl_with_arg_pattern.search(text) or tmpl_no_arg_pattern.search(text):
            match_result = tmpl_with_arg_pattern.search(text) or tmpl_no_arg_pattern.search(text)
            template_define = match_result.group(0)
            self._log_replace_process(f"found template: {template_define}")
            template_define_normalized = template_define.replace("\n", "")
            template_name = match_result.groups()[0]
            if template_name not in templates:
                raise SqlProcessorException(f"no template for found `{template_name}`, existing are {templates}")

            template = templates.get(template_name)
            values = re.compile(r"\s*\w+\s*=\s*[^,)]+,?\s*", flags=re.IGNORECASE).findall(template_define_normalized)
            if values:
                index = 0
                while index < len(values):
                    value_def = str(values[index]).split("=")
                    value_name = value_def[0].strip()
                    value = value_def[1].replace(",", "").strip()
                    # fix for the last template parameter
                    if re.compile(r"\)}$").search(value):
                        value = value[:-2].strip()
                    self._log_replace_process(
                        f"template param matched: value_name={value_name}, value: {value}, template_name:"
                        f" {template_name}, template: {template}"
                    )
                    template = re.sub(re.escape(f"#{{{value_name}}}"), value, template, flags=re.IGNORECASE)
                    index += 1
            text = re.sub(re.escape(template_define), template, text, flags=re.IGNORECASE)
            self._log_replace_process(f"text after template replaced: {text}")

        return text

    def _log_replace_process(self, message: str):
        if self.debug_log:
            logger.debug(message)

    def add_templates(self, templates: Dict[str, str]):
        templates = {k.lower(): v for k, v in templates.items()}
        self.templates.update(templates)
        if self.debug_log:
            logger.debug(f"templates: {self.templates}")


class ProcessorContext:
    def __init__(self, vars_context: VarsContext, templates_context: TemplatesContext, extra_cols: List[Column] = None):
        self.vars_context = vars_context
        self.templates_context = templates_context
        self.extra_cols = extra_cols or []

    def replace_templates(self, text: str):
        return self.templates_context.replace_templates(text)

    def replace_variables(self, text: str) -> str:
        return self.vars_context.replace_variables(text)

    @property
    def vars(self):
        return self.vars_context.vars

    def set_vars(self, vars: Dict[str, Any]):
        self.vars_context.vars = vars

    def add_vars(self, vars: Dict[str, Any]):
        self.vars_context.add_vars(vars)

    def add_list_vars(self, list_vars: Dict[str, List]):
        self.vars_context.add_list_vars(list_vars)

    def add_templates(self, templates: Dict[str, str]):
        self.templates_context.add_templates(templates)



from sqlfluff.core.parser import CodeSegment
from sqlfluff.core.rules.base import BaseRule, LintFix, LintResult, RuleContext


class Rule_BigQuery_L001(BaseRule):
    """
    Table schema is required for queries in BigQuery.

    **Anti-pattern**
    Select from some table without schema.

    .. code-block:: sql
        SELECT *
        FROM foo

    **Best practice**
    Select from some table with schema.

    .. code-block:: sql
        SELECT *
        FROM test.foo
    """

    groups = ("all", "bigquery")

    def __init__(self, *args, **kwargs):
        """Overwrite __init__ to set config."""
        super().__init__(*args, **kwargs)

    def _eval(self, context: RuleContext):
        """check from table have schema"""
        if context.segment.is_type("table_reference") and len(context.segment.segments) != 3:
            return LintResult(
                anchor=context.segment,
                fixes=[
                    LintFix.create_before(
                        context.segment,
                        [CodeSegment(raw="${temp_db}.")],
                    )
                ],
                description=f"No schema found when select from table `{context.segment.raw}`.",
            )
